# Copyright (c) Alibaba, Inc. and its affiliates.
import random
from typing import Dict, List, Tuple

import numpy as np
import torch
import torch.nn as nn
from mmcv.cnn import kaiming_init, normal_init
from mmcv.runner import get_dist_info
from pytorch_metric_learning.miners import *
from torch import Tensor

from easycv.core.evaluation.metrics import accuracy
from easycv.models.loss import CrossEntropyLossWithLabelSmooth
from easycv.models.utils import DistributedLossWrapper, DistributedMinerWrapper
from easycv.utils.logger import get_root_logger
from easycv.utils.registry import build_from_cfg
from ..registry import HEADS, LOSSES

# Softmax based loss doesn't need ddp, the big fc while slowdown the training process.
MP_NODDP_LOSS = set([
    'ArcFaceLoss', 'AngularLoss', 'CosFaceLoss', 'LargeMarginSoftmaxLoss',
    'NormalizedSoftmaxLoss', 'SphereFaceLoss',
    'CrossEntropyLossWithLabelSmooth', 'AMSoftmaxLoss'
])


def EmbeddingExplansion(embs, labels, explanion_rate=4, alpha=1.0):
    """
        Expand embedding: CVPR refer to  https://github.com/clovaai/embedding-expansion
        combine PK sampled data, mixup anchor positive pair to generate more features, always combine with BatchHardminer.
        result on SOP and CUB need to be add

        Args:
            embs: [N , dims] tensor
            labels: [N] tensor
            explanion_rate: to expand N to explanion_rate * N
            alpha: beta distribution parameter for mixup

        Return:
            embs: [N*explanion_rate , dims]
    """

    embs = embs[0]
    label_list = labels.cpu().data.numpy()
    label_keys = []
    label_idx = {}

    # caculate label and mixup
    old = -1
    for idx, i in enumerate(label_list):
        if i == old or i in label_idx.keys():
            label_idx[old].append(idx)
        else:
            label_idx[i] = [idx]
            old = i
            label_keys.append(old)

    res = embs
    res_label = labels

    for j in range(explanion_rate - 1):
        refine_label_list = []
        for key in label_keys:
            random.shuffle(label_idx[key])
            refine_label_list += label_idx[key]

        refine_label_list = torch.tensor(refine_label_list).to(embs.device)

        if alpha > 0:
            lam = np.random.beta(
                alpha, alpha, size=[refine_label_list.size(0)])
        else:
            lam = 1

        lam = torch.tensor(lam).view(len(refine_label_list),
                                     -1).to(embs.device)

        data_mixed = lam * embs + (1 - lam) * embs[refine_label_list, :]
        data_mixed = data_mixed.float()
        res = torch.cat([res, data_mixed])
        res_label = torch.cat([res_label, labels])

    return [res], res_label


@HEADS.register_module
class MpMetrixHead(nn.Module):
    """Simplest classifier head, with only one fc layer.
    """

    def __init__(
        self,
        with_avg_pool=False,
        in_channels=2048,
        loss_config=[{
            'type': 'CircleLoss',
            'loss_weight': 1.0,
            'norm': True,
            'ddp': True,
            'm': 0.4,
            'gamma': 80
        }],
        input_feature_index=[0],
        input_label_index=0,
        ignore_label=None,
    ):

        super(MpMetrixHead, self).__init__()
        self.with_avg_pool = with_avg_pool
        self.in_channels = in_channels
        self.input_feature_index = input_feature_index
        self.input_label_index = input_label_index
        self.ignore_label = ignore_label

        rank, world_size = get_dist_info()

        logger = get_root_logger()
        if self.with_avg_pool:
            self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))

        self.loss_list = []
        self.norm_list = []
        self.loss_weight_list = []
        self.ddp_list = []
        self.miner_list = []
        assert len(loss_config) > 0
        for idx, loss in enumerate(loss_config):
            self.loss_weight_list.append(loss.pop('loss_weight', 1.0))
            self.norm_list.append(loss.pop('norm', True))
            cbm_param = loss.pop('cbm', None)
            miner_param = loss.pop('miner', None)
            name = loss['type']

            # ddp will be True is user not set and name not in MP_NODDP_LOSS
            ddp = loss.pop('ddp', None)
            if ddp is None:
                if name in MP_NODDP_LOSS:
                    ddp = False
                else:
                    ddp = True
            self.ddp_list.append(ddp)

            if world_size > 1 and self.ddp_list[idx]:
                tmp = build_from_cfg(loss, LOSSES)
                tmp_loss = DistributedLossWrapper(loss=tmp)
            else:
                tmp_loss = build_from_cfg(loss, LOSSES)

            if miner_param is not None:
                name = miner_param.pop('type')
                if world_size > 1 and self.ddp_list[idx]:
                    self.miner_list.append(
                        DistributedMinerWrapper(eval(name)(**miner_param)))
                else:
                    self.miner_list.append(eval(name)(**miner_param))
            else:
                self.miner_list.append(None)

            setattr(self, '%s_%d' % (name, idx), tmp_loss)
            self.loss_list.append(getattr(self, '%s_%d' % (name, idx)))

    def init_weights(self,
                     pretrained=None,
                     init_linear='normal',
                     std=0.01,
                     bias=0.):
        assert init_linear in ['normal', 'kaiming'], \
            'Undefined init_linear: {}'.format(init_linear)
        for m in self.modules():
            if isinstance(m, nn.Linear):
                if init_linear == 'normal':
                    normal_init(m, std=std, bias=bias)
                else:
                    kaiming_init(m, mode='fan_in', nonlinearity='relu')
            elif isinstance(m,
                            (nn.BatchNorm2d, nn.GroupNorm, nn.SyncBatchNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def forward(self, x: List[torch.Tensor]) -> List[torch.Tensor]:
        # multi head feature distribute
        for i in self.input_feature_index:
            assert i < len(x)
        x = [x[i] for i in self.input_feature_index]
        assert isinstance(x, (tuple, list)) and len(x) == 1

        x1 = x[0]
        if self.with_avg_pool and x1.dim() > 2:
            assert x1.dim() == 4, \
                'Tensor must has 4 dims, got: {}'.format(x.dim())
            x1 = self.avg_pool(x1)
        x1 = x1.view(x1.size(0), -1)

        if hasattr(self, 'fc_cls'):
            cls_score = self.fc_cls(x1)
        else:
            cls_score = x1

        return [cls_score]

    def loss(self, cls_score, labels) -> Dict[str, torch.Tensor]:
        logger = get_root_logger()

        losses = dict()
        assert isinstance(cls_score, (tuple, list)) and len(cls_score) == 1

        if type(labels) == list:
            assert (self.input_label_index < len(labels))
            tlabel = labels[self.input_label_index]
        else:
            tlabel = labels

        if self.ignore_label is not None:
            ignore_mask = tlabel.eq(self.ignore_label)
            no_ignore_mask = ~ignore_mask
            tlabel = torch.masked_select(tlabel, no_ignore_mask)
            no_ignore_idx = torch.where(no_ignore_mask == True)[0]
            cls_score = [
                torch.index_select(tcls, 0, no_ignore_idx)
                for tcls in cls_score
            ]

        loss = None

        for i in range(0, len(self.norm_list)):
            if self.norm_list[i]:
                a = torch.nn.functional.normalize(cls_score[0], p=2, dim=1)
            else:
                a = cls_score[0]

            if self.miner_list[i] is not None:
                tuple_indice = self.miner_list[i](a, tlabel)
                if not torch.isnan(self.loss_list[i](a, tlabel, tuple_indice)):
                    if loss is None:
                        loss = self.loss_weight_list[i] * self.loss_list[i](
                            a, tlabel, tuple_indice)
                    else:
                        loss += self.loss_weight_list[i] * self.loss_list[i](
                            a, tlabel, tuple_indice)
                else:
                    logger.info(
                        'MP metric head catch NAN loss in %dth loss !' % i)
            else:
                if not torch.isnan(self.loss_list[i](a, tlabel)):
                    if loss is None:
                        loss = self.loss_weight_list[i] * self.loss_list[i](
                            a, tlabel)
                    else:
                        loss += self.loss_weight_list[i] * self.loss_list[i](
                            a, tlabel)
                else:
                    logger.info(
                        'MP metric head catch NAN loss in %dth loss !' % i)

        if loss is None:
            loss = torch.tensor(
                0.0, requires_grad=True).to(cls_score[0].device)

        losses['loss'] = loss
        try:
            losses['acc'] = accuracy(a, tlabel)
        except:
            pass

        return losses



# Copyright (c) Alibaba, Inc. and its affiliates.
import numpy as np

from easycv.datasets.registry import PIPELINES


@PIPELINES.register_module()
class DaliImageDecoder(object):
    """refer to: https://docs.nvidia.com/deeplearning/dali/archives/dali_0250/user-guide/docs/supported_ops.html#nvidia.dali.ops.ImageDecoder
    """

    def __init__(self, device='mixed', **kwargs):
        import nvidia.dali.ops as ops

        self.decode_op = ops.ImageDecoder(device=device, **kwargs)

    def __call__(self, images):
        return self.decode_op(images)


@PIPELINES.register_module()
class DaliRandomResizedCrop(object):
    """refer to: https://docs.nvidia.com/deeplearning/dali/archives/dali_0250/user-guide/docs/supported_ops.html#nvidia.dali.ops.RandomResizedCrop
    """

    def __init__(self, size, random_area, device='gpu', **kwargs):
        import nvidia.dali.ops as ops

        self.random_resize_crop_op = ops.RandomResizedCrop(
            size=size, random_area=random_area, device=device, **kwargs)

    def __call__(self, images, **kwargs):
        return self.random_resize_crop_op(images, **kwargs)


@PIPELINES.register_module()
class DaliResize(object):
    """refer to: https://docs.nvidia.com/deeplearning/dali/archives/dali_0250/user-guide/docs/supported_ops.html#nvidia.dali.ops.Resize
    """

    def __init__(self, resize_shorter, device='gpu', **kwargs):
        import nvidia.dali.ops as ops
        import nvidia.dali.types as types

        _kwargs = dict(interp_type=types.INTERP_TRIANGULAR)
        _kwargs.update(kwargs)
        self.resize_op = ops.Resize(
            device=device, resize_shorter=resize_shorter, **_kwargs)

    def __call__(self, images, **kwargs):
        return self.resize_op(images, **kwargs)


@PIPELINES.register_module()
class DaliColorTwist(object):
    """refer to: https://docs.nvidia.com/deeplearning/dali/archives/dali_0250/user-guide/docs/supported_ops.html#nvidia.dali.ops.ColorTwist
    """

    def __init__(self,
                 prob,
                 saturation,
                 contrast,
                 brightness,
                 hue,
                 device='gpu',
                 center=1):
        import nvidia.dali.ops as ops
        import nvidia.dali.types as types

        self.color_twist_op = ops.ColorTwist(device=device)
        self.saturation_op = ops.Uniform(
            range=[center - saturation, center + saturation])
        self.contrast_op = ops.Uniform(
            range=[center - contrast, center + contrast])
        self.brightness_op = ops.Uniform(
            range=[center - brightness, center + brightness])
        self.hue_op = ops.Uniform(range=[-hue, hue])

        self.coin_flip_op = ops.CoinFlip(probability=prob)
        self.cast_fp32_op = ops.Cast(dtype=types.FLOAT)

    def __call__(self, images):
        # saturation: default = 1.0 no change
        # brightness: default = 1.0 no change
        # contrast: default = 1.0 no change
        # hue: default = 0.0 no change

        coin_flip = self.cast_fp32_op(self.coin_flip_op())

        saturation = ((1 - coin_flip) * 1.0 + coin_flip * self.saturation_op())
        contrast = ((1 - coin_flip) * 1.0 + coin_flip * self.contrast_op())
        brightness = ((1 - coin_flip) * 1.0 + coin_flip * self.brightness_op())
        hue = coin_flip * self.hue_op()

        output = self.color_twist_op(
            images,
            saturation=saturation,
            contrast=contrast,
            brightness=brightness,
            hue=hue)
        return output


@PIPELINES.register_module()
class DaliRandomGrayscale(object):
    """refer to: https://docs.nvidia.com/deeplearning/dali/archives/dali_0250/user-guide/docs/supported_ops.html#nvidia.dali.ops.Hsv
    Create `RandomGrayscale` op with ops.Hsv.
    when saturation=0, it represents a grayscale image
    """

    def __init__(self, prob, device='gpu'):
        import nvidia.dali.ops as ops
        import nvidia.dali.types as types

        self.coin_flip_op = ops.CoinFlip(probability=1 - prob)
        self.cast_fp32_op = ops.Cast(dtype=types.FLOAT)
        self.hsv_op = ops.Hsv(device=device)

    def __call__(self, images):
        saturation = self.cast_fp32_op(self.coin_flip_op())
        output = self.hsv_op(images, saturation=saturation)

        return output


@PIPELINES.register_module()
class DaliCropMirrorNormalize(object):
    """refer to: https://docs.nvidia.com/deeplearning/dali/archives/dali_0250/user-guide/docs/supported_ops.html#nvidia.dali.ops.CropMirrorNormalize
    """

    def __init__(self,
                 crop,
                 mean,
                 std,
                 prob=0.0,
                 device='gpu',
                 crop_pos_x=0.5,
                 crop_pos_y=0.5,
                 **kwargs):
        import nvidia.dali.ops as ops
        import nvidia.dali.types as types

        _kwargs = dict(dtype=types.FLOAT, output_layout=types.NCHW)
        _kwargs.update(kwargs)

        if isinstance(crop, int):
            crop = [crop, crop]

        # TODO: support random crop_pos by iter
        # We designed crop_pos_x and crop_pos_y random by epoch
        # we has reproduced the accuracy by epoch
        if isinstance(crop_pos_x, (tuple, list)):
            assert len(crop_pos_x) == 2
            crop_pos_x = np.random.uniform(
                low=crop_pos_x[0], high=crop_pos_x[1])
        if isinstance(crop_pos_y, (tuple, list)):
            assert len(crop_pos_y) == 2
            crop_pos_y = np.random.uniform(
                low=crop_pos_y[0], high=crop_pos_y[1])

        self.device = device
        self.coin_flip_op = ops.CoinFlip(probability=prob)
        self.cmn_op = ops.CropMirrorNormalize(
            device=device,
            crop=crop,
            # image_type=image_type,
            crop_pos_x=crop_pos_x,
            crop_pos_y=crop_pos_y,
            mean=mean,
            std=std,
            **_kwargs)

    def __call__(self, images, **kwargs):
        if self.device == 'gpu':
            images = images.gpu()

        output = self.cmn_op(
            images,
            mirror=self.coin_flip_op(),  # flipped (mirrored) horizontally.
            **kwargs)

        return output



_base_ = 'configs/classification/imagenet/inception/inceptionv3_b32x8_100e.py'

num_classes = 1000
# model settings
model = dict(
    type='Classification',
    backbone=dict(type='Inception4', num_classes=num_classes),
    head=[
        dict(
            type='ClsHead',
            with_fc=False,
            in_channels=1536,
            loss_config=dict(
                type='CrossEntropyLossWithLabelSmooth',
                label_smooth=0,
            ),
            num_classes=num_classes,
            input_feature_index=[1],
        ),
        dict(
            type='ClsHead',
            with_fc=False,
            in_channels=768,
            loss_config=dict(
                type='CrossEntropyLossWithLabelSmooth',
                label_smooth=0,
            ),
            num_classes=num_classes,
            input_feature_index=[0],
        )
    ])

img_norm_cfg = dict(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])



_base_ = 'configs/base.py'

log_config = dict(
    interval=200,
    hooks=[
        dict(type='TextLoggerHook'),
        # dict(type='TensorboardLoggerHook')
    ])

checkpoint_config = dict(interval=10)

# optimizer
optimizer = dict(
    type='AdamW',
    lr=1e-4,
    betas=(0.9, 0.999),
    weight_decay=0.1,
    constructor='LayerDecayOptimizerConstructor',
    paramwise_options=dict(num_layers=12, layer_decay_rate=0.7))
optimizer_config = dict(grad_clip=None)
# learning policy
lr_config = dict(
    policy='step',
    warmup='linear',
    warmup_iters=250,
    warmup_ratio=0.001,
    step=[88, 96])
total_epochs = 100

find_unused_parameters = False



# Copyright (c) Alibaba, Inc. and its affiliates.
import inspect
import logging

from modelscope.trainers.hooks import HOOKS
from modelscope.trainers.parallel.builder import PARALLEL
from modelscope.utils.registry import default_group


class _RegisterManager:

    def __init__(self):
        self.registries = {}

    def add(self, module, name, group_key=default_group):
        if module.name not in self.registries:
            self.registries[module.name] = {}
        if group_key not in self.registries[module.name]:
            self.registries[module.name][group_key] = []

        self.registries[module.name][group_key].append(name)

    def exists(self, module, name, group_key=default_group):
        if self.registries.get(module.name, None) is None:
            return False
        if self.registries[module.name].get(group_key, None) is None:
            return False
        if name in self.registries[module.name][group_key]:
            return True

        return False


_dynamic_register = _RegisterManager()


def register_parallel():
    from mmcv.parallel import MMDistributedDataParallel, MMDataParallel

    mmddp = 'MMDistributedDataParallel'
    mmdp = 'MMDataParallel'

    if not _dynamic_register.exists(PARALLEL, mmddp):
        _dynamic_register.add(PARALLEL, mmddp)
        PARALLEL.register_module(
            module_name=mmddp, module_cls=MMDistributedDataParallel)
    if not _dynamic_register.exists(PARALLEL, mmdp):
        _dynamic_register.add(PARALLEL, mmdp)
        PARALLEL.register_module(module_name=mmdp, module_cls=MMDataParallel)


def register_hook_to_ms(hook_name, logger=None):
    """Register EasyCV hook to ModelScope."""
    from easycv.hooks import HOOKS as _EV_HOOKS

    if hook_name not in _EV_HOOKS._module_dict:
        raise ValueError(
            f'Not found hook "{hook_name}" in EasyCV hook registries!')

    if _dynamic_register.exists(HOOKS, hook_name):
        return
    _dynamic_register.add(HOOKS, hook_name)

    obj = _EV_HOOKS._module_dict[hook_name]
    HOOKS.register_module(module_name=hook_name, module_cls=obj)

    log_str = f'Register hook "{hook_name}" to modelscope hooks.'
    logger.info(log_str) if logger is not None else logging.info(log_str)


def register_part_mmcv_hooks_to_ms():
    """Register required mmcv hooks to ModelScope.
    Currently we only registered all lr scheduler hooks in EasyCV and mmcv.
    Please refer to:
        EasyCV: https://github.com/alibaba/EasyCV/blob/master/easycv/hooks/lr_update_hook.py
        mmcv: https://github.com/open-mmlab/mmcv/blob/master/mmcv/runner/hooks/lr_updater.py
    """
    from mmcv.runner.hooks import lr_updater
    from mmcv.runner.hooks import HOOKS as _MMCV_HOOKS
    from easycv.hooks import StepFixCosineAnnealingLrUpdaterHook, YOLOXLrUpdaterHook

    mmcv_hooks_in_easycv = [('StepFixCosineAnnealingLrUpdaterHook',
                             StepFixCosineAnnealingLrUpdaterHook),
                            ('YOLOXLrUpdaterHook', YOLOXLrUpdaterHook)]

    members = inspect.getmembers(lr_updater)
    members.extend(mmcv_hooks_in_easycv)

    for name, obj in members:
        if name in _MMCV_HOOKS._module_dict:
            if _dynamic_register.exists(HOOKS, name):
                continue
            _dynamic_register.add(HOOKS, name)
            HOOKS.register_module(
                module_name=name,
                module_cls=obj,
            )



# Copyright (c) Alibaba, Inc. and its affiliates.

import cv2

from easycv.predictors.builder import PREDICTORS
from .base import OutputProcessor, PredictorV2

face_contour_point_index = [
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
    21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32
]
left_eye_brow_point_index = [33, 34, 35, 36, 37, 38, 39, 40, 41, 33]
right_eye_brow_point_index = [42, 43, 44, 45, 46, 47, 48, 49, 50, 42]
left_eye_point_index = [66, 67, 68, 69, 70, 71, 72, 73, 66]
right_eye_point_index = [75, 76, 77, 78, 79, 80, 81, 82, 75]
nose_bridge_point_index = [51, 52, 53, 54]
nose_contour_point_index = [55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65]
mouth_outer_point_index = [84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 84]
mouth_inter_point_index = [96, 97, 98, 99, 100, 101, 102, 103, 96]


class FaceKptsOutputProcessor(OutputProcessor):
    """Process the output of face keypoints models.

    Args:
        input_size (int): Target image size.
    """

    def __init__(self, input_size, point_number):
        self.input_size = input_size
        self.point_number = point_number

    def __call__(self, inputs):
        results = []

        img_metas = inputs['img_metas']
        points = inputs['point'].cpu().numpy()
        poses = inputs['pose'].cpu().numpy()

        for idx, point in enumerate(points):
            h, w, c = img_metas[idx]['img_shape']
            scale_h = h / self.input_size
            scale_w = w / self.input_size

            point = point.reshape((self.point_number, 2))
            for index in range(len(point)):
                point[index][0] *= scale_w
                point[index][1] *= scale_h

            results.append({'point': point, 'pose': poses[idx]})

        return results


@PREDICTORS.register_module()
class FaceKeypointsPredictor(PredictorV2):
    """Predict pipeline for face keypoint
    Args:
        model_path (str): Path of model path
        config_file (str): config file path for model and processor to init. Defaults to None.
        batch_size (int): batch size for forward.
        device (str): Support 'cuda' or 'cpu', if is None, detect device automatically.
        save_results (bool): Whether to save predict results.
        save_path (str): File path for saving results, only valid when `save_results` is True.
        pipelines (list[dict]): Data pipeline configs.
        input_processor_threads (int): Number of processes to process inputs.
        mode (str): The image mode into the model.
    """

    def __init__(
        self,
        model_path,
        config_file,
        batch_size=1,
        device=None,
        save_results=False,
        save_path=None,
        pipelines=None,
        input_processor_threads=8,
        mode='BGR',
    ):
        super(FaceKeypointsPredictor, self).__init__(
            model_path,
            config_file,
            batch_size=batch_size,
            device=device,
            save_results=save_results,
            save_path=save_path,
            pipelines=pipelines,
            input_processor_threads=input_processor_threads,
            mode=mode)

        self.input_size = self.cfg.IMAGE_SIZE
        self.point_number = self.cfg.POINT_NUMBER

    def model_forward(self, inputs):
        outputs = super().model_forward(inputs)
        outputs['img_metas'] = inputs['img_metas']
        return outputs

    def get_output_processor(self):
        return FaceKptsOutputProcessor(
            input_size=self.input_size, point_number=self.point_number)

    def show_result(self, img, points, scale=4.0, save_path=None):
        """Draw `result` over `img`.

        Args:
            img ( ndarray ): The image to be displayed.
            result (list): The face keypoints to draw over `img`.
            scale: zoom in or out scale
            save_path: path to save drawned 'img'
        Returns:
            img (ndarray): Only if not `show` or `out_file`
        """

        image = cv2.resize(img, dsize=None, fx=scale, fy=scale)

        def draw_line(point_index, image, point):
            for i in range(len(point_index) - 1):
                cur_index = point_index[i]
                next_index = point_index[i + 1]
                cur_pt = (int(point[cur_index][0] * scale),
                          int(point[cur_index][1] * scale))
                next_pt = (int(point[next_index][0] * scale),
                           int(point[next_index][1] * scale))
                cv2.line(image, cur_pt, next_pt, (0, 0, 255), thickness=2)

        draw_line(face_contour_point_index, image, points)
        draw_line(left_eye_brow_point_index, image, points)
        draw_line(right_eye_brow_point_index, image, points)
        draw_line(left_eye_point_index, image, points)
        draw_line(right_eye_point_index, image, points)
        draw_line(nose_bridge_point_index, image, points)
        draw_line(nose_contour_point_index, image, points)
        draw_line(mouth_outer_point_index, image, points)
        draw_line(mouth_inter_point_index, image, points)

        size = len(points)
        for i in range(size):
            x = int(points[i][0])
            y = int(points[i][1])
            cv2.putText(image, str(i), (int(x * scale), int(y * scale)),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 255, 0), 1)
            cv2.circle(image, (int(x * scale), int(y * scale)), 2, (0, 255, 0),
                       cv2.FILLED)

        if save_path is not None:
            cv2.imwrite(save_path, image)

        return image



# Copyright (c) Alibaba, Inc. and its affiliates.
import argparse
import inspect
from collections import Counter

import numpy as np
import tqdm
from fvcore.nn import FlopCountAnalysis, flop_count_table
from mmcv.parallel import scatter_kwargs
from prettytable import PrettyTable

from easycv.datasets.builder import build_dataset
from easycv.datasets.loader import build_dataloader
from easycv.datasets.utils import is_dali_dataset_type
from easycv.models.builder import build_model
from easycv.utils.config_tools import mmcv_config_fromfile
from easycv.utils.mmlab_utils import dynamic_adapt_for_mmlab


def parse_args():
    parser = argparse.ArgumentParser(description='count model flops')
    parser.add_argument('config', help='config file path')
    parser.add_argument(
        '--repeat_num', default=10, type=int, help='repeat number')
    args = parser.parse_args()
    return args


def flatten_inputs(model, inputs):
    full_args_spec = inspect.getfullargspec(model.forward)
    args = [] if not full_args_spec.args else full_args_spec.args
    args.pop(0) if (args and args[0] in ['self', 'cls']) else args

    default_values = [] if not full_args_spec.defaults else full_args_spec.defaults
    args_has_default = args[len(args) - len(default_values):]
    args_with_default = dict(zip(
        args_has_default, default_values)) if len(args_has_default) else {}

    flat_inputs = []
    for arg_i in args:
        if inputs.get(arg_i, None) is not None:
            flat_inputs.append(inputs.get(arg_i))
        else:
            flat_inputs.append(args_with_default.get(arg_i))

    return tuple(flat_inputs)


def count_flop():
    args = parse_args()

    device = 'cuda'
    cfg = mmcv_config_fromfile(args.config)

    # dynamic adapt mmdet models
    dynamic_adapt_for_mmlab(cfg)

    model = build_model(cfg.model)
    model.to(device)
    model.eval()

    if cfg.data.get('val', None) is not None:
        cfg.data.val.pop('imgs_per_gpu', None)  # pop useless params
        data_cfg = cfg.data.val
    else:
        data_cfg = cfg.data.train

    if is_dali_dataset_type(data_cfg['type']):
        data_cfg.distributed = False
        data_cfg.batch_size = 1
        data_cfg.workers_per_gpu = 1
        dataset = build_dataset(data_cfg)
        data_loader = dataset.get_dataloader()
    else:
        dataset = build_dataset(data_cfg)
        data_loader = build_dataloader(
            dataset, imgs_per_gpu=1, workers_per_gpu=0)

    handlers = {}  # mapping from operator names to handles.
    counts = Counter()
    gflop_unit = 1e9
    total_flops = []
    for idx, data in zip(tqdm.trange(args.repeat_num), data_loader):
        # use scatter_kwargs to unpack DataContainer data for raw torch.nn.module
        _, kwargs = scatter_kwargs(None, data, [0])
        kwargs[0].update({'mode': 'test'})
        inputs = flatten_inputs(model, kwargs[0])

        # Provides access to per-submodule model flop count obtained by
        # tracing a model with pytorch's jit tracing functionality.
        # So models that donot support jit tracing may fail.
        flops = FlopCountAnalysis(model, inputs)
        flops.set_op_handle(**handlers)
        if idx > 0:
            flops.unsupported_ops_warnings(False).uncalled_modules_warnings(
                False)
        counts += flops.by_operator()
        total_flops.append(flops.total())

    print('Flops from only one sample is:\n' + flop_count_table(flops))
    ops_show = PrettyTable()
    ops_show.field_names = ['operator type', 'Gflops']
    for k, v in counts.items():
        ops_show.add_row([k, round(v / (idx + 1) / gflop_unit, 3)])
    print('Average Gflops of each operator type is:')
    print(ops_show)
    print('Total flops: {:.1f}G ± {:.1f}G'.format(
        np.mean(total_flops) / gflop_unit,
        np.std(total_flops) / gflop_unit))


if __name__ == '__main__':
    count_flop()



# Copyright (c) Alibaba, Inc. and its affiliates.
from easycv.utils.registry import Registry, build_from_cfg

EVALUATORS = Registry('hook')


def build_evaluator(evaluator_cfg_list):
    """ build evaluator according to metric name

    Args:
        evaluator_cfg_list: list of evaluator config dict
    Return:
        return a list of evaluator
    """
    if isinstance(evaluator_cfg_list, dict):
        evaluator_cfg_list = [evaluator_cfg_list]
    evaluators = []

    for cfg in evaluator_cfg_list:
        evaluators.append(build_from_cfg(cfg, EVALUATORS))

    return evaluators



# Copyright (c) Alibaba, Inc. and its affiliates.
import math
from glob import glob

import numpy as np
import torch
from PIL import Image

from easycv.framework.errors import ValueError
from easycv.thirdparty.face_align import glint360k_align
from easycv.thirdparty.mtcnn import FaceDetector
from .base import Predictor
from .builder import PREDICTORS

try:
    from easy_vision.python.inference.predictor import PredictorInterface
except:
    from .interface import PredictorInterface


@PREDICTORS.register_module()
class TorchFeatureExtractor(PredictorInterface):

    def __init__(self, model_path, model_config=None):
        """
    init model

    Args:
      model_path: model file path
      model_config: config string for model to init, in json format
    """
        self.predictor = Predictor(model_path)
        self.output_name = 'feature'

    def get_output_type(self):
        """
    in this function user should return a type dict, which indicates
    which type of data should the output of predictor be converted to
    * type json, data will be serialized to json str

    * type image, data will be converted to encode image binary and write to oss file,
      whose name is output_dir/${key}/${input_filename}_${idx}.jpg, where input_filename
      is the base filename extracted from url, key corresponds to the key in the dict of output_type,
      if the type of data indexed by key is a list, idx is the index of element in list, otherwhile ${idx} will be empty

    * type video, data will be converted to encode video binary and write to oss file,

    :: return  {
      'image': 'image',
      'feature': 'json'
    }
    indicating that the image data in the output dict will be save to image
    file and feature in output dict will be converted to json

    """
        return {}

    def batch(self, image_tensor_list):
        return torch.stack(image_tensor_list)

    def predict(self, input_data_list, batch_size=-1):
        """
    using session run predict a number of samples using batch_size

    Args:
      input_data_list:  a list of numpy array, each array is a sample to be predicted
      batch_size: batch_size passed by the caller, you can also ignore this param and
        use a fixed number if you do not want to adjust batch_size in runtime
    Return:
      result: a list of dict, each dict is the prediction result of one sample
        eg, {"output1": value1, "output2": value2}, the value type can be
        python int str float, and numpy array
    """
        num_image = len(input_data_list)
        assert len(
            input_data_list) > 0, 'input images should not be an empty list'
        if batch_size > 0:
            num_batches = int(math.ceil(float(num_image) / batch_size))
            image_list = input_data_list
        else:
            num_batches = 1
            batch_size = len(input_data_list)
            image_list = input_data_list

        outputs_list = []
        for batch_idx in range(num_batches):
            batch_image_list = image_list[batch_idx * batch_size:min(
                (batch_idx + 1) * batch_size, len(image_list))]
            image_tensor_list = self.predictor.preprocess(batch_image_list)
            input_data = self.batch(image_tensor_list)
            outputs = self.predictor.predict_batch(
                input_data, mode='extract')['neck'].data.cpu().numpy()
            for idx in range(len(image_tensor_list)):
                single_result = {}
                out = np.squeeze(outputs[idx])
                single_result[self.output_name] = out
                outputs_list.append(single_result)

        return outputs_list


@PREDICTORS.register_module()
class TorchFaceFeatureExtractor(PredictorInterface):

    def __init__(self, model_path, model_config=None):
        """
    init model, add a facedetect and align for img input.

    Args:
      model_path: model file path
      model_config: config string for model to init, in json format
    """

        # Forward compatibility, to support both pth(only face feature model) or tar.gz(face feature model + mtcnn model)
        if model_path.endswith('.pth') or model_path.endswith('.pt'):
            self.predictor = Predictor(model_path)
            self.detector = FaceDetector()
        else:
            face_model = glob('%s/*.pth' % model_path) + glob(
                '%s/*.pt' % model_path)
            assert (len(face_model) == 1)
            self.predictor = Predictor(face_model[0])

            mtcnn_weights = glob('%s/weights/*.npy' % model_path)
            if len(mtcnn_weights) != 3:
                print(
                    "User provide model_path doesn't contain mtcnn models, we try to load weights from http, might failed!"
                )
            self.detector = FaceDetector(dir_path=model_path)

        self.output_name = 'feature'

    def get_output_type(self):
        """
    in this function user should return a type dict, which indicates
    which type of data should the output of predictor be converted to
    * type json, data will be serialized to json str

    * type image, data will be converted to encode image binary and write to oss file,
      whose name is output_dir/${key}/${input_filename}_${idx}.jpg, where input_filename
      is the base filename extracted from url, key corresponds to the key in the dict of output_type,
      if the type of data indexed by key is a list, idx is the index of element in list, otherwhile ${idx} will be empty

    * type video, data will be converted to encode video binary and write to oss file,

    :: return  {
      'image': 'image',
      'feature': 'json'
    }
    indicating that the image data in the output dict will be save to image
    file and feature in output dict will be converted to json

    """
        return {}

    def batch(self, image_tensor_list):
        return torch.stack(image_tensor_list)

    def predict(self, input_data_list, batch_size=-1, detect_and_align=True):
        """
    using session run predict a number of samples using batch_size

    Args:
      input_data_list:  a list of numpy array or PIL.Image, each array is a sample to be predicted
      batch_size: batch_size passed by the caller, you can also ignore this param and
        use a fixed number if you do not want to adjust batch_size in runtime
      detect_and_align: True to detect and align before feature extractor
    Return:
      result: a list of dict, each dict is the prediction result of one sample
        eg, {"output1": value1, "output2": value2}, the value type can be
        python int str float, and numpy array
    Raise:
       if detect !=1 face in a img, then do nothing for this image
    """
        num_image = len(input_data_list)
        assert len(
            input_data_list) > 0, 'input images should not be an empty list'
        if batch_size > 0:
            num_batches = int(math.ceil(float(num_image) / batch_size))
            image_list = input_data_list
        else:
            num_batches = 1
            batch_size = len(input_data_list)
            image_list = input_data_list

        for i in range(len(image_list)):
            if isinstance(image_list[i], np.ndarray):
                image_list[i] = Image.fromarray(image_list[i])

        outputs_list = []
        for batch_idx in range(num_batches):
            batch_image_list = image_list[batch_idx * batch_size:min(
                (batch_idx + 1) * batch_size, len(image_list))]
            if detect_and_align:
                for idx, img in enumerate(batch_image_list):
                    bbox, ld = self.detector.safe_detect(img)
                    if len(bbox) > 1:
                        print('batch %d , %dth image has more then 1 face' %
                              (batch_idx, idx))
                        batch_image_list[idx] = np.array(
                            glint360k_align(img, ld[0]))
                    elif len(bbox) == 0:
                        print(
                            'batch %d , %dth image has no face detected, use original img'
                            % (batch_idx, idx))
                        batch_image_list[idx] = np.array(
                            img.resize((112, 112)))
                    else:
                        batch_image_list[idx] = np.array(
                            glint360k_align(img, ld[0]))
            else:
                for idx, img in enumerate(batch_image_list):
                    batch_image_list[idx] = np.array(img.resize((112, 112)))

            image_tensor_list = self.predictor.preprocess(batch_image_list)
            input_data = self.batch(image_tensor_list)
            outputs = self.predictor.predict_batch(
                input_data, mode='extract')['neck'].data.cpu().numpy()

            for idx in range(len(image_tensor_list)):
                single_result = {}
                out = np.squeeze(outputs[idx])
                single_result[self.output_name] = out
                outputs_list.append(single_result)

        return outputs_list


@PREDICTORS.register_module()
class TorchMultiFaceFeatureExtractor(PredictorInterface):

    def __init__(self, model_path, model_config=None):
        """
    init model, add a facedetect and align for img input.

    Args:
      model_path: model file path
      model_config: config string for model to init, in json format
    """

        # Forward compatibility, to support both pth(only face feature model) or tar.gz(face feature model + mtcnn model)
        if model_path.endswith('.pth') or model_path.endswith('.pt'):
            self.predictor = Predictor(model_path)
            self.detector = FaceDetector()
        else:
            face_model = glob('%s/*.pth' % model_path) + glob(
                '%s/*.pt' % model_path)
            assert (len(face_model) == 1)
            self.predictor = Predictor(face_model[0])

            mtcnn_weights = glob('%s/weights/*.npy' % model_path)
            if len(mtcnn_weights) != 3:
                print(
                    "User provide model_path doesn't contain mtcnn models, we try to load weights from http, might failed!"
                )
            self.detector = FaceDetector(dir_path=model_path)

        self.output_name = 'feature'

    def get_output_type(self):
        """
    in this function user should return a type dict, which indicates
    which type of data should the output of predictor be converted to
    * type json, data will be serialized to json str

    * type image, data will be converted to encode image binary and write to oss file,
      whose name is output_dir/${key}/${input_filename}_${idx}.jpg, where input_filename
      is the base filename extracted from url, key corresponds to the key in the dict of output_type,
      if the type of data indexed by key is a list, idx is the index of element in list, otherwhile ${idx} will be empty

    * type video, data will be converted to encode video binary and write to oss file,

    :: return  {
      'image': 'image',
      'feature': 'json'
    }
    indicating that the image data in the output dict will be save to image
    file and feature in output dict will be converted to json

    """
        return {}

    def batch(self, image_tensor_list):
        return torch.stack(image_tensor_list)

    def predict(self, input_data_list, batch_size=-1, detect_and_align=True):
        """
    using session run predict a number of samples using batch_size

    Args:
      input_data_list:  a list of numpy array or PIL.Image, each array is a sample to be predicted
      batch_size: batch_size passed by the caller, you can also ignore this param and
        use a fixed number if you do not want to adjust batch_size in runtime
      detect_and_align: True to detect and align before feature extractor
    Return:
      result: a list of dict, each dict is the prediction result of one sample
        eg, {"output1": value1, "output2": value2}, the value type can be
        python int str float, and numpy array
    Raise:
       if detect !=1 face in a img, then do nothing for this image
    """
        num_image = len(input_data_list)

        outputs_list = []
        for idx in range(num_image):
            img = input_data_list[idx]
            if isinstance(img, np.ndarray):
                img = Image.fromarray(img)
            sub_imgs = []
            if detect_and_align:
                bboxes, ld = self.detector.safe_detect(img)
                if len(bboxes) > 1:
                    print('%dth image has more then 1 face' % idx)
                    for box, subld in zip(bboxes, ld):
                        sub_imgs.append(np.array(glint360k_align(img, subld)))
                    # sub_imgs.append(np.array(glint360k_align(img, ld[0])))
                elif len(bboxes) == 0:
                    print('%dth image has no face detected, use original img' %
                          idx)
                    sub_imgs.append(np.array(img.resize((112, 112))))
                else:
                    sub_imgs.append(np.array(glint360k_align(img, ld[0])))
            else:
                sub_imgs.append(np.array(img.resize((112, 112))))
                # x1,y1 x2,y2,score
                bboxes = [[0, 0, 111, 111, 1.0]]

            image_tensor_list = self.predictor.preprocess(sub_imgs)
            input_data = self.batch(image_tensor_list)
            outputs = self.predictor.predict_batch(
                input_data, mode='extract')['neck'].data.cpu().numpy()

            # for sub_idx in range(len(image_tensor_list)):
            single_result = {}
            if len(bboxes) == 0:
                bboxes = [[0, 0, 111, 111, 1.0]]

            # out = np.squeeze(outputs[0])

            single_result[self.output_name] = outputs
            single_result['bbox'] = bboxes

            outputs_list.append(single_result)

        return outputs_list


@PREDICTORS.register_module()
class TorchFaceAttrExtractor(PredictorInterface):

    def __init__(
        self,
        model_path,
        model_config=None,
        face_threshold=0.95,
        attr_method=['distribute_sum', 'softmax', 'softmax'],
        attr_name=['age', 'gender', 'emo'],
    ):
        """
    init model

    Args:
      model_path: model file path
      model_config: config string for model to init, in json format
      attr_method:
        - softmax: do softmax for feature_dim 1
        - distribute_sum: do softmax and prob sum
    """
        if model_path.endswith('.pth') or model_path.endswith('.pt'):
            self.predictor = Predictor(model_path)
            self.detector = FaceDetector()
        else:
            face_model = glob('%s/*.pth' % model_path) + glob(
                '%s/*.pt' % model_path)
            assert (len(face_model) == 1)
            self.predictor = Predictor(face_model[0])

            mtcnn_weights = glob('%s/weights/*.npy' % model_path)
            if len(mtcnn_weights) != 3:
                print(
                    "User provide model_path doesn't contain mtcnn models, we try to load weights from http, might failed!"
                )
            self.detector = FaceDetector(dir_path=model_path)

        self.attr_name = attr_name
        self.attr_method = attr_method
        assert (len(self.attr_method) == len(self.attr_name))
        self.gender_map = {0: 'female', 1: 'male'}
        self.emo_map = {
            0: 'Neutral',
            1: 'Happiness',
            2: 'Sadness',
            3: 'Surprise',
            4: 'Fear',
            5: 'Disgust',
            6: 'Anger',
            7: 'Contempt',
        }
        self.pop_map = {
            0: '0',
            1: '1',
            2: '2',
            3: '3',
            4: '4',
            5: '5',
        }
        self.face_threshold = face_threshold

    def get_output_type(self):
        """
    in this function user should return a type dict, which indicates
    which type of data should the output of predictor be converted to
    * type json, data will be serialized to json str

    * type image, data will be converted to encode image binary and write to oss file,
      whose name is output_dir/${key}/${input_filename}_${idx}.jpg, where input_filename
      is the base filename extracted from url, key corresponds to the key in the dict of output_type,
      if the type of data indexed by key is a list, idx is the index of element in list, otherwhile ${idx} will be empty

    * type video, data will be converted to encode video binary and write to oss file,

    :: return  {
      'image': 'image',
      'feature': 'json'
    }
    indicating that the image data in the output dict will be save to image
    file and feature in output dict will be converted to json

    """
        return {}

    def batch(self, image_tensor_list):
        return torch.stack(image_tensor_list)

    def predict(self, input_data_list, batch_size=-1):
        """
    using session run predict a number of samples using batch_size

    Args:
      input_data_list:  a list of numpy array, each array is a sample to be predicted
      batch_size: batch_size passed by the caller, you can also ignore this param and
        use a fixed number if you do not want to adjust batch_size in runtime
    Return:
      result: a list of dict, each dict is the prediction result of one sample
        eg, {"output1": value1, "output2": value2}, the value type can be
        python int str float, and numpy array
    """
        num_image = len(input_data_list)
        assert len(
            input_data_list) > 0, 'input images should not be an empty list'
        if batch_size > 0:
            num_batches = int(math.ceil(float(num_image) / batch_size))
            image_list = input_data_list
        else:
            num_batches = 1
            batch_size = len(input_data_list)
            image_list = input_data_list

        outputs_list = []
        for batch_idx in range(num_batches):
            batch_image_list = image_list[
                batch_idx *
                batch_size:min(len(image_list), (batch_idx + 1) * batch_size)]

            face_image_list = []
            face_bbox_list = []
            faceidx_by_imageidx = {}

            for idx, img in enumerate(batch_image_list):
                # this try except only happens to no face detected
                bbox, ld = self.detector.safe_detect(img)

                if len(bbox) == 0:
                    print('batch %d , %dth image has no face detected' %
                          (batch_idx, idx))
                elif len(bbox) >= 1:
                    if len(bbox) > 1:
                        print('batch %d , %dth image has more then %d face' %
                              (batch_idx, idx, len(bbox)))

                    _bbox = []
                    _ld = []
                    for idx, b in enumerate(bbox):
                        if b[-1] > self.face_threshold:
                            _bbox.append(b)
                            _ld.append(ld[idx])

                    bbox = _bbox
                    ld = _ld

                    # this is for muti face detectd in one img
                    faceidx_by_imageidx[idx] = []
                    for bbox_idx, face_box in enumerate(bbox):
                        face_image_list.append(
                            glint360k_align(img, ld[bbox_idx]))
                        face_bbox_list.append(face_box)
                        face_idx = len(face_image_list) - 1
                        faceidx_by_imageidx[idx].append(face_idx)
                # else:
                #     batch_image_list[idx] = np.array(glint360k_align(img, ld[0]))

            if len(face_image_list) > 0:
                image_tensor_list = self.predictor.preprocess(face_image_list)
                input_data = self.batch(image_tensor_list)

                outputs = self.predictor.predict_batch(
                    input_data, mode='extract')

                neck_output_dict = {}
                for neck_idx, attr_method in enumerate(self.attr_method):
                    neck_output = outputs['neck_%d_0' % neck_idx]
                    neck_output = torch.nn.Softmax(dim=1)(neck_output)
                    if attr_method == 'softmax':
                        neck_output = torch.argmax(neck_output, dim=1)
                    elif attr_method == 'distribute_sum':
                        n, c = neck_output.size()
                        distribute = torch.arange(0, c).repeat(n, 1).to(
                            neck_output.device)
                        neck_output = (distribute * neck_output).sum(dim=1)
                    else:
                        raise ValueError(
                            'TorchFaceAttrExtractor for neck %d only support attr_method softmax/distributed sum'
                            % (neck_idx))
                        neck_output = torch.argmax(neck_output, dim=1)
                    neck_output_dict[neck_idx] = neck_output.cpu().numpy()

                for imgidx in faceidx_by_imageidx.keys():
                    single_result = {}
                    for k in neck_output_dict.keys():
                        single_result['face_' + self.attr_name[k]] = []
                    single_result['face_bbox'] = []
                    for fn, faceidx in enumerate(faceidx_by_imageidx[imgidx]):
                        for k in neck_output_dict.keys():
                            out = np.squeeze(neck_output_dict[k][faceidx])
                            if self.attr_method[k] == 'softmax':
                                label_map = getattr(
                                    self, '%s_map' % self.attr_name[k])
                                out = label_map[out]
                            single_result['face_' +
                                          self.attr_name[k]].append(out)
                        single_result['face_bbox'].append(
                            face_bbox_list[faceidx])
                    outputs_list.append(single_result)
        return outputs_list



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import math
import os.path as osp

import cv2
import numpy as np
import requests

from easycv.datasets.registry import PIPELINES
from easycv.utils.logger import get_root_logger


@PIPELINES.register_module()
class BaseRecLabelEncode(object):
    """ Convert between text-label and text-index """

    def __init__(self,
                 max_text_length,
                 character_dict_path=None,
                 use_space_char=False):

        self.max_text_len = max_text_length
        self.BEGIN_STR = 'sos'
        self.END_STR = 'eos'
        self.lower = False

        if character_dict_path is None:
            logger = get_root_logger()
            logger.warning(
                'The character_dict_path is None, model can only recognize number and lower letters'
            )
            self.character_str = '0123456789abcdefghijklmnopqrstuvwxyz'
            dict_character = list(self.character_str)
            self.lower = True
        else:
            self.character_str = []
            if character_dict_path.startswith('http'):
                r = requests.get(character_dict_path)
                tpath = character_dict_path.split('/')[-1]
                while not osp.exists(tpath):
                    try:
                        with open(tpath, 'wb') as code:
                            code.write(r.content)
                    except:
                        pass
                character_dict_path = tpath
            with open(character_dict_path, 'rb') as fin:
                lines = fin.readlines()
                for line in lines:
                    line = line.decode('utf-8').strip('\n').strip('\r\n')
                    self.character_str.append(line)
            if use_space_char:
                self.character_str.append(' ')
            dict_character = list(self.character_str)
        dict_character = self.add_special_char(dict_character)
        self.dict = {}
        for i, char in enumerate(dict_character):
            self.dict[char] = i
        self.character = dict_character

    def add_special_char(self, dict_character):
        return dict_character

    def encode(self, text):
        """convert text-label into text-index.
        input:
            text: text labels of each image. [batch_size]

        output:
            text: concatenated text index for CTCLoss.
                    [sum(text_lengths)] = [text_index_0 + text_index_1 + ... + text_index_(n - 1)]
            length: length of each text. [batch_size]
        """
        if len(text) == 0 or len(text) > self.max_text_len:
            return None
        if self.lower:
            text = text.lower()
        text_list = []
        for char in text:
            if char not in self.dict:
                # logger = get_logger()
                # logger.warning('{} is not in dict'.format(char))
                continue
            text_list.append(self.dict[char])
        if len(text_list) == 0:
            return None
        return text_list


@PIPELINES.register_module()
class CTCLabelEncode(BaseRecLabelEncode):
    """ Convert between text-label and text-index """

    def __init__(self,
                 max_text_length,
                 character_dict_path=None,
                 use_space_char=False,
                 **kwargs):
        self.BLANK = ['blank']
        super(CTCLabelEncode,
              self).__init__(max_text_length, character_dict_path,
                             use_space_char)

    def __call__(self, data):
        text = data['label']
        text = self.encode(text)
        if text is None:
            return None
        data['length'] = np.array(len(text))
        text = text + [0] * (self.max_text_len - len(text))
        data['label'] = np.array(text)

        label = [0] * len(self.character)
        for x in text:
            label[x] += 1
        data['label_ace'] = np.array(label)
        return data

    def add_special_char(self, dict_character):
        dict_character = self.BLANK + dict_character
        return dict_character


@PIPELINES.register_module()
class SARLabelEncode(BaseRecLabelEncode):
    """ Convert between text-label and text-index """

    def __init__(self,
                 max_text_length,
                 character_dict_path=None,
                 use_space_char=False,
                 **kwargs):
        self.BEG_END_STR = '<BOS/EOS>'
        self.UNKNOWN_STR = '<UKN>'
        self.PADDING_STR = '<PAD>'
        super(SARLabelEncode,
              self).__init__(max_text_length, character_dict_path,
                             use_space_char)

    def add_special_char(self, dict_character):
        dict_character = dict_character + [self.UNKNOWN_STR]
        self.unknown_idx = len(dict_character) - 1
        dict_character = dict_character + [self.BEG_END_STR]
        self.start_idx = len(dict_character) - 1
        self.end_idx = len(dict_character) - 1
        dict_character = dict_character + [self.PADDING_STR]
        self.padding_idx = len(dict_character) - 1

        return dict_character

    def __call__(self, data):
        text = data['label']
        text = self.encode(text)
        if text is None:
            return None
        if len(text) >= self.max_text_len - 1:
            return None
        data['length'] = np.array(len(text))
        target = [self.start_idx] + text + [self.end_idx]
        padded_text = [self.padding_idx for _ in range(self.max_text_len)]

        padded_text[:len(target)] = target
        data['label'] = np.array(padded_text)
        return data

    def get_ignored_tokens(self):
        return [self.padding_idx]


@PIPELINES.register_module()
class MultiLabelEncode(BaseRecLabelEncode):

    def __init__(self,
                 max_text_length,
                 character_dict_path=None,
                 use_space_char=False,
                 **kwargs):
        super(MultiLabelEncode,
              self).__init__(max_text_length, character_dict_path,
                             use_space_char)

        self.ctc_encode = CTCLabelEncode(max_text_length, character_dict_path,
                                         use_space_char, **kwargs)
        self.sar_encode = SARLabelEncode(max_text_length, character_dict_path,
                                         use_space_char, **kwargs)

    def __call__(self, data):

        data_ctc = copy.deepcopy(data)
        data_sar = copy.deepcopy(data)
        data_out = dict()
        data_out['img_path'] = data.get('img_path', None)
        data_out['img'] = data['img']
        ctc = self.ctc_encode(data_ctc)
        sar = self.sar_encode(data_sar)
        if ctc is None or sar is None:
            return None
        data_out['label_ctc'] = ctc['label']
        data_out['label_sar'] = sar['label']
        data_out['length'] = ctc['length']
        return data_out



# Copyright (c) Alibaba, Inc. and its affiliates.
# Copyright (c) OpenMMLab. All rights reserved.
# Refer to: https://github.com/open-mmlab/mmaction2/blob/master/mmaction/models/backbones/stgcn.py
import torch
import torch.nn as nn
from mmcv.cnn import constant_init, kaiming_init, normal_init
from mmcv.utils import _BatchNorm

from easycv.models import BACKBONES
from easycv.utils.checkpoint import load_checkpoint
from easycv.utils.logger import get_root_logger
from .utils import Graph


def zero(x):
    """return zero."""
    return 0


def identity(x):
    """return input itself."""
    return x


class STGCNBlock(nn.Module):
    """Applies a spatial temporal graph convolution over an input graph
    sequence.

    Args:
        in_channels (int): Number of channels in the input sequence data
        out_channels (int): Number of channels produced by the convolution
        kernel_size (tuple): Size of the temporal convolving kernel and
            graph convolving kernel
        stride (int, optional): Stride of the temporal convolution. Default: 1
        dropout (int, optional): Dropout rate of the final output. Default: 0
        residual (bool, optional): If ``True``, applies a residual mechanism.
            Default: ``True``

    Shape:
        - Input[0]: Input graph sequence in :math:`(N, in_channels, T_{in}, V)`
            format
        - Input[1]: Input graph adjacency matrix in :math:`(K, V, V)` format
        - Output[0]: Outpu graph sequence in :math:`(N, out_channels, T_{out},
            V)` format
        - Output[1]: Graph adjacency matrix for output data in :math:`(K, V,
            V)` format

        where
            :math:`N` is a batch size,
            :math:`K` is the spatial kernel size, as :math:`K == kernel_size[1]
                `,
            :math:`T_{in}/T_{out}` is a length of input/output sequence,
            :math:`V` is the number of graph nodes.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride=1,
                 dropout=0,
                 residual=True):
        super().__init__()

        assert len(kernel_size) == 2
        assert kernel_size[0] % 2 == 1
        padding = ((kernel_size[0] - 1) // 2, 0)

        self.gcn = ConvTemporalGraphical(in_channels, out_channels,
                                         kernel_size[1])
        self.tcn = nn.Sequential(
            nn.BatchNorm2d(out_channels), nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, (kernel_size[0], 1),
                      (stride, 1), padding), nn.BatchNorm2d(out_channels),
            nn.Dropout(dropout, inplace=True))

        if not residual:
            self.residual = zero

        elif (in_channels == out_channels) and (stride == 1):
            self.residual = identity

        else:
            self.residual = nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size=1,
                    stride=(stride, 1)), nn.BatchNorm2d(out_channels))

        self.relu = nn.ReLU(inplace=True)

    def forward(self, x, adj_mat):
        """Defines the computation performed at every call."""
        res = self.residual(x)
        x, adj_mat = self.gcn(x, adj_mat)
        x = self.tcn(x) + res

        return self.relu(x), adj_mat


class ConvTemporalGraphical(nn.Module):
    """The basic module for applying a graph convolution.

    Args:
        in_channels (int): Number of channels in the input sequence data
        out_channels (int): Number of channels produced by the convolution
        kernel_size (int): Size of the graph convolving kernel
        t_kernel_size (int): Size of the temporal convolving kernel
        t_stride (int, optional): Stride of the temporal convolution.
            Default: 1
        t_padding (int, optional): Temporal zero-padding added to both sides
            of the input. Default: 0
        t_dilation (int, optional): Spacing between temporal kernel elements.
            Default: 1
        bias (bool, optional): If ``True``, adds a learnable bias to the
            output. Default: ``True``

    Shape:
        - Input[0]: Input graph sequence in :math:`(N, in_channels, T_{in}, V)`
            format
        - Input[1]: Input graph adjacency matrix in :math:`(K, V, V)` format
        - Output[0]: Output graph sequence in :math:`(N, out_channels, T_{out}
            , V)` format
        - Output[1]: Graph adjacency matrix for output data in :math:`(K, V, V)
            ` format

        where
            :math:`N` is a batch size,
            :math:`K` is the spatial kernel size, as :math:`K == kernel_size[1]
                `,
            :math:`T_{in}/T_{out}` is a length of input/output sequence,
            :math:`V` is the number of graph nodes.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 t_kernel_size=1,
                 t_stride=1,
                 t_padding=0,
                 t_dilation=1,
                 bias=True):
        super().__init__()

        self.kernel_size = kernel_size
        self.conv = nn.Conv2d(
            in_channels,
            out_channels * kernel_size,
            kernel_size=(t_kernel_size, 1),
            padding=(t_padding, 0),
            stride=(t_stride, 1),
            dilation=(t_dilation, 1),
            bias=bias)

    def forward(self, x, adj_mat):
        """Defines the computation performed at every call."""
        assert adj_mat.size(0) == self.kernel_size

        x = self.conv(x)

        n, kc, t, v = x.size()
        x = x.view(n, self.kernel_size, kc // self.kernel_size, t, v)
        x = torch.einsum('nkctv,kvw->nctw', (x, adj_mat))

        return x.contiguous(), adj_mat


@BACKBONES.register_module()
class STGCN(nn.Module):
    """Backbone of Spatial temporal graph convolutional networks.

    Args:
        in_channels (int): Number of channels in the input data.
        graph_cfg (dict): The arguments for building the graph.
        edge_importance_weighting (bool): If ``True``, adds a learnable
            importance weighting to the edges of the graph. Default: True.
        data_bn (bool): If 'True', adds data normalization to the inputs.
            Default: True.
        pretrained (str | None): Name of pretrained model.
        **kwargs (optional): Other parameters for graph convolution units.

    Shape:
        - Input: :math:`(N, in_channels, T_{in}, V_{in}, M_{in})`
        - Output: :math:`(N, num_class)` where
            :math:`N` is a batch size,
            :math:`T_{in}` is a length of input sequence,
            :math:`V_{in}` is the number of graph nodes,
            :math:`M_{in}` is the number of instance in a frame.
    """

    def __init__(self,
                 in_channels,
                 graph_cfg,
                 edge_importance_weighting=True,
                 data_bn=True,
                 pretrained=None,
                 **kwargs):
        super().__init__()

        # load graph
        self.graph = Graph(**graph_cfg)
        A = torch.tensor(
            self.graph.A, dtype=torch.float32, requires_grad=False)
        self.register_buffer('A', A)

        # build networks
        spatial_kernel_size = A.size(0)
        temporal_kernel_size = 9
        kernel_size = (temporal_kernel_size, spatial_kernel_size)
        self.data_bn = nn.BatchNorm1d(in_channels *
                                      A.size(1)) if data_bn else identity

        kwargs0 = {k: v for k, v in kwargs.items() if k != 'dropout'}
        self.st_gcn_networks = nn.ModuleList((
            STGCNBlock(
                in_channels, 64, kernel_size, 1, residual=False, **kwargs0),
            STGCNBlock(64, 64, kernel_size, 1, **kwargs),
            STGCNBlock(64, 64, kernel_size, 1, **kwargs),
            STGCNBlock(64, 64, kernel_size, 1, **kwargs),
            STGCNBlock(64, 128, kernel_size, 2, **kwargs),
            STGCNBlock(128, 128, kernel_size, 1, **kwargs),
            STGCNBlock(128, 128, kernel_size, 1, **kwargs),
            STGCNBlock(128, 256, kernel_size, 2, **kwargs),
            STGCNBlock(256, 256, kernel_size, 1, **kwargs),
            STGCNBlock(256, 256, kernel_size, 1, **kwargs),
        ))

        # initialize parameters for edge importance weighting
        if edge_importance_weighting:
            self.edge_importance = nn.ParameterList([
                nn.Parameter(torch.ones(self.A.size()))
                for i in self.st_gcn_networks
            ])
        else:
            self.edge_importance = [1 for _ in self.st_gcn_networks]

        self.pretrained = pretrained

    def init_weights(self):
        """Initiate the parameters either from existing checkpoint or from
        scratch."""
        if isinstance(self.pretrained, str):
            logger = get_root_logger()
            logger.info(f'load model from: {self.pretrained}')

            load_checkpoint(self, self.pretrained, strict=False, logger=logger)

        elif self.pretrained is None:
            for m in self.modules():
                if isinstance(m, nn.Conv2d):
                    kaiming_init(m)
                elif isinstance(m, nn.Linear):
                    normal_init(m)
                elif isinstance(m, _BatchNorm):
                    constant_init(m, 1)
        else:
            raise TypeError('pretrained must be a str or None')

    def forward(self, x):
        """Defines the computation performed at every call.
        Args:
            x (torch.Tensor): The input data.

        Returns:
            torch.Tensor: The output of the module.
        """
        # data normalization
        x = x.float()
        n, c, t, v, m = x.size()  # bs 3 300 25(17) 2
        x = x.permute(0, 4, 3, 1, 2).contiguous()  # N M V C T
        x = x.view(n * m, v * c, t)
        x = self.data_bn(x)
        x = x.view(n, m, v, c, t)
        x = x.permute(0, 1, 3, 4, 2).contiguous()
        x = x.view(n * m, c, t, v)  # bsx2 3 300 25(17)

        # forward
        for gcn, importance in zip(self.st_gcn_networks, self.edge_importance):
            x, _ = gcn(x, self.A * importance)

        return x



# Copyright (c) Alibaba, Inc. and its affiliates.
def generate_best_metric_name(evaluate_type, dataset_name, metric_names):
    """
    Generate best metric name for different evaluator / different dataset / different metric_names
    evaluate_type: str
    dataset_name: None or str
    metric_names: None str or list[str] or tuple(str)

    Return:
        list[str]
    """
    base_name = evaluate_type
    if dataset_name is not None:
        base_name = base_name + '_' + dataset_name

    return_name = []
    if metric_names is None:
        return_name = [base_name]
    elif type(metric_names) == str:
        return_name = [base_name + '_' + metric_names]
    elif type(metric_names) == list or type(metric_names) == tuple:
        if len(metric_names) == 0:
            return_name = [base_name]
        for k in metric_names:
            assert (type(k) == str)
            return_name.append(base_name + '_' + k)

    return return_name



from __future__ import absolute_import, division, print_function
from collections import namedtuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import constant_init, kaiming_init
from torch.nn.modules.batchnorm import _BatchNorm

from ..modelzoo import inceptionv4 as model_urls
from ..registry import BACKBONES

__all__ = ['Inception4']


class BasicConv2d(nn.Module):

    def __init__(self,
                 in_planes,
                 out_planes,
                 kernel_size,
                 stride=1,
                 padding=0):
        super(BasicConv2d, self).__init__()
        self.conv = nn.Conv2d(
            in_planes,
            out_planes,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=False)  # verify bias false
        self.bn = nn.BatchNorm2d(
            out_planes,
            eps=0.001,  # value found in tensorflow
            momentum=0.1,  # default pytorch value
            affine=True)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class Mixed_3a(nn.Module):

    def __init__(self):
        super(Mixed_3a, self).__init__()
        self.maxpool = nn.MaxPool2d(3, stride=2)
        self.conv = BasicConv2d(64, 96, kernel_size=3, stride=2)

    def forward(self, x):
        x0 = self.maxpool(x)
        x1 = self.conv(x)
        out = torch.cat((x0, x1), 1)
        return out


class Mixed_4a(nn.Module):

    def __init__(self):
        super(Mixed_4a, self).__init__()

        self.branch0 = nn.Sequential(
            BasicConv2d(160, 64, kernel_size=1, stride=1),
            BasicConv2d(64, 96, kernel_size=3, stride=1))

        self.branch1 = nn.Sequential(
            BasicConv2d(160, 64, kernel_size=1, stride=1),
            BasicConv2d(64, 64, kernel_size=(1, 7), stride=1, padding=(0, 3)),
            BasicConv2d(64, 64, kernel_size=(7, 1), stride=1, padding=(3, 0)),
            BasicConv2d(64, 96, kernel_size=(3, 3), stride=1))

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        out = torch.cat((x0, x1), 1)
        return out


class Mixed_5a(nn.Module):

    def __init__(self):
        super(Mixed_5a, self).__init__()
        self.conv = BasicConv2d(192, 192, kernel_size=3, stride=2)
        self.maxpool = nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        x0 = self.conv(x)
        x1 = self.maxpool(x)
        out = torch.cat((x0, x1), 1)
        return out


class Inception_A(nn.Module):

    def __init__(self):
        super(Inception_A, self).__init__()
        self.branch0 = BasicConv2d(384, 96, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(384, 64, kernel_size=1, stride=1),
            BasicConv2d(64, 96, kernel_size=3, stride=1, padding=1))

        self.branch2 = nn.Sequential(
            BasicConv2d(384, 64, kernel_size=1, stride=1),
            BasicConv2d(64, 96, kernel_size=3, stride=1, padding=1),
            BasicConv2d(96, 96, kernel_size=3, stride=1, padding=1))

        self.branch3 = nn.Sequential(
            nn.AvgPool2d(3, stride=1, padding=1, count_include_pad=False),
            BasicConv2d(384, 96, kernel_size=1, stride=1))

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = self.branch3(x)
        out = torch.cat((x0, x1, x2, x3), 1)
        return out


class Reduction_A(nn.Module):

    def __init__(self):
        super(Reduction_A, self).__init__()
        self.branch0 = BasicConv2d(384, 384, kernel_size=3, stride=2)

        self.branch1 = nn.Sequential(
            BasicConv2d(384, 192, kernel_size=1, stride=1),
            BasicConv2d(192, 224, kernel_size=3, stride=1, padding=1),
            BasicConv2d(224, 256, kernel_size=3, stride=2))

        self.branch2 = nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        out = torch.cat((x0, x1, x2), 1)
        return out


class Inception_B(nn.Module):

    def __init__(self):
        super(Inception_B, self).__init__()
        self.branch0 = BasicConv2d(1024, 384, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(1024, 192, kernel_size=1, stride=1),
            BasicConv2d(
                192, 224, kernel_size=(1, 7), stride=1, padding=(0, 3)),
            BasicConv2d(
                224, 256, kernel_size=(7, 1), stride=1, padding=(3, 0)))

        self.branch2 = nn.Sequential(
            BasicConv2d(1024, 192, kernel_size=1, stride=1),
            BasicConv2d(
                192, 192, kernel_size=(7, 1), stride=1, padding=(3, 0)),
            BasicConv2d(
                192, 224, kernel_size=(1, 7), stride=1, padding=(0, 3)),
            BasicConv2d(
                224, 224, kernel_size=(7, 1), stride=1, padding=(3, 0)),
            BasicConv2d(
                224, 256, kernel_size=(1, 7), stride=1, padding=(0, 3)))

        self.branch3 = nn.Sequential(
            nn.AvgPool2d(3, stride=1, padding=1, count_include_pad=False),
            BasicConv2d(1024, 128, kernel_size=1, stride=1))

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = self.branch3(x)
        out = torch.cat((x0, x1, x2, x3), 1)
        return out


class Reduction_B(nn.Module):

    def __init__(self):
        super(Reduction_B, self).__init__()

        self.branch0 = nn.Sequential(
            BasicConv2d(1024, 192, kernel_size=1, stride=1),
            BasicConv2d(192, 192, kernel_size=3, stride=2))

        self.branch1 = nn.Sequential(
            BasicConv2d(1024, 256, kernel_size=1, stride=1),
            BasicConv2d(
                256, 256, kernel_size=(1, 7), stride=1, padding=(0, 3)),
            BasicConv2d(
                256, 320, kernel_size=(7, 1), stride=1, padding=(3, 0)),
            BasicConv2d(320, 320, kernel_size=3, stride=2))

        self.branch2 = nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        out = torch.cat((x0, x1, x2), 1)
        return out


class Inception_C(nn.Module):

    def __init__(self):
        super(Inception_C, self).__init__()

        self.branch0 = BasicConv2d(1536, 256, kernel_size=1, stride=1)

        self.branch1_0 = BasicConv2d(1536, 384, kernel_size=1, stride=1)
        self.branch1_1a = BasicConv2d(
            384, 256, kernel_size=(1, 3), stride=1, padding=(0, 1))
        self.branch1_1b = BasicConv2d(
            384, 256, kernel_size=(3, 1), stride=1, padding=(1, 0))

        self.branch2_0 = BasicConv2d(1536, 384, kernel_size=1, stride=1)
        self.branch2_1 = BasicConv2d(
            384, 448, kernel_size=(3, 1), stride=1, padding=(1, 0))
        self.branch2_2 = BasicConv2d(
            448, 512, kernel_size=(1, 3), stride=1, padding=(0, 1))
        self.branch2_3a = BasicConv2d(
            512, 256, kernel_size=(1, 3), stride=1, padding=(0, 1))
        self.branch2_3b = BasicConv2d(
            512, 256, kernel_size=(3, 1), stride=1, padding=(1, 0))

        self.branch3 = nn.Sequential(
            nn.AvgPool2d(3, stride=1, padding=1, count_include_pad=False),
            BasicConv2d(1536, 256, kernel_size=1, stride=1))

    def forward(self, x):
        x0 = self.branch0(x)

        x1_0 = self.branch1_0(x)
        x1_1a = self.branch1_1a(x1_0)
        x1_1b = self.branch1_1b(x1_0)
        x1 = torch.cat((x1_1a, x1_1b), 1)

        x2_0 = self.branch2_0(x)
        x2_1 = self.branch2_1(x2_0)
        x2_2 = self.branch2_2(x2_1)
        x2_3a = self.branch2_3a(x2_2)
        x2_3b = self.branch2_3b(x2_2)
        x2 = torch.cat((x2_3a, x2_3b), 1)

        x3 = self.branch3(x)

        out = torch.cat((x0, x1, x2, x3), 1)
        return out


class InceptionAux(nn.Module):

    def __init__(self, in_channels, num_classes):
        super(InceptionAux, self).__init__()
        self.conv0 = BasicConv2d(in_channels, 128, kernel_size=1)
        self.conv1 = BasicConv2d(128, 768, kernel_size=5)
        self.conv1.stddev = 0.01
        self.fc = nn.Linear(768, num_classes)
        self.fc.stddev = 0.001

    def forward(self, x):
        # N x 768 x 17 x 17
        x = F.avg_pool2d(x, kernel_size=5, stride=3)
        # N x 768 x 5 x 5
        x = self.conv0(x)
        # N x 128 x 5 x 5
        x = self.conv1(x)
        # N x 768 x 1 x 1
        # Adaptive average pooling
        x = F.adaptive_avg_pool2d(x, (1, 1))
        # N x 768 x 1 x 1
        x = torch.flatten(x, 1)
        # N x 768
        x = self.fc(x)
        # N x 1000
        return x


# class BasicConv2d(nn.Module):

#     def __init__(self, in_channels, out_channels, **kwargs):
#         super(BasicConv2d, self).__init__()
#         self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)
#         self.bn = nn.BatchNorm2d(out_channels, eps=0.001)

#     def forward(self, x):
#         x = self.conv(x)
#         x = self.bn(x)
#         return F.relu(x, inplace=True)


@BACKBONES.register_module
class Inception4(nn.Module):
    """InceptionV4 backbone.

    Args:
        num_classes (int): The num_classes of InceptionV4. An extra fc will be used if
    """

    def __init__(self,
                 num_classes: int = 0,
                 p_dropout=0.2,
                 aux_logits: bool = True):
        super(Inception4, self).__init__()
        self.aux_logits = aux_logits
        # Modules
        self.features = nn.Sequential(
            BasicConv2d(3, 32, kernel_size=3, stride=2),
            BasicConv2d(32, 32, kernel_size=3, stride=1),
            BasicConv2d(32, 64, kernel_size=3, stride=1, padding=1),
            Mixed_3a(),
            Mixed_4a(),
            Mixed_5a(),
            Inception_A(),
            Inception_A(),
            Inception_A(),
            Inception_A(),
            Reduction_A(),  # Mixed_6a
            Inception_B(),
            Inception_B(),
            Inception_B(),
            Inception_B(),
            Inception_B(),
            Inception_B(),
            Inception_B(),  # Mixed_6h 1024 x 17 x 17
            Reduction_B(),  # Mixed_7a
            Inception_C(),
            Inception_C(),
            Inception_C())

        if aux_logits:
            self.AuxLogits = InceptionAux(1024, num_classes)

        self.dropout = nn.Dropout(p_dropout)
        self.last_linear = None
        if num_classes > 0:
            self.last_linear = nn.Linear(1536, num_classes)

        self.default_pretrained_model_path = model_urls[
            self.__class__.__name__]

    @property
    def fc(self):
        return self.last_linear

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                for m in self.modules():
                    if isinstance(m, nn.Conv2d):
                        kaiming_init(m, mode='fan_in', nonlinearity='relu')
                    elif isinstance(m, (_BatchNorm, nn.GroupNorm)):
                        constant_init(m, 1)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def logits(self, features):
        x = F.adaptive_avg_pool2d(features, output_size=(1, 1))
        # x = F.avg_pool2d(features, kernel_size=adaptiveAvgPoolWidth)
        x = x.view(x.size(0), -1)  # B x 1536
        x = self.fc(x)
        # B x num_classes
        return x

    def forward(self, input: torch.Tensor):
        """_summary_

        Args:
            input (torch.Tensor): A RGB image tensor with shape B x C x H x W

        Returns:
            torch.Tensor: A feature tensor or a logit tensor when num_classes is 0 (default)
        """

        if self.training and self.aux_logits:
            x = self.features[:-4](input)
            aux = self.AuxLogits(x)
            x = self.features[-4:](x)
        else:
            x = self.features(input)
            aux = None

        if self.fc is not None:
            x = self.logits(x)

        return [aux, x]



# Copyright (c) Alibaba, Inc. and its affiliates.
import configparser
import logging
import os
import sys
import urllib
from collections import namedtuple
from contextlib import contextmanager
from io import StringIO

from tqdm import tqdm

from easycv.framework.errors import ValueError

OSS_PREFIX = 'oss://'
URL_PREFIX = ('https://', 'http://')


def create_namedtuple(**kwargs):
    return namedtuple('namedtuple', list(kwargs.keys()))(**kwargs)


def is_oss_path(s):
    return s.startswith(OSS_PREFIX)


def is_url_path(s):
    return s.startswith(URL_PREFIX)


def url_path_exists(url):
    try:
        urllib.request.urlopen(url).code
    except Exception as err:
        print(err)
        return False
    return True


def get_oss_config():
    """
    Get oss config file from env `OSS_CONFIG_FILE`,
    default file is `~/.ossutilconfig`.
    """
    oss_cfg_file = os.environ.get('OSS_CONFIG_FILE', '~/.ossutilconfig')
    oss_cfg_file = os.path.expanduser(oss_cfg_file)

    if not os.path.isabs(oss_cfg_file):
        raise ValueError('Not support relative path for `OSS_CONFIG_FILE`!')

    if not os.path.exists(oss_cfg_file):
        raise ValueError(
            'Please add the oss config file and add `[Bucket-Endpoint]` in your oss config file, '
            'refer to: https://help.aliyun.com/document_detail/120072.html')

    cfg_parser = configparser.ConfigParser()
    cfg_parser.read(oss_cfg_file)

    if 'Bucket-Endpoint' not in cfg_parser:
        raise ValueError(
            'Please add `[Bucket-Endpoint]` in your oss config file, refer to: https://help.aliyun.com/document_detail/120072.html'
        )

    credential = dict(cfg_parser['Credentials'])
    bucket_endpoint = dict(cfg_parser['Bucket-Endpoint'])
    buckets = list(bucket_endpoint.keys())
    hosts = list(bucket_endpoint.values())

    oss_config = dict(
        ak_id=credential['accesskeyid'],
        ak_secret=credential['accesskeysecret'],
        hosts=hosts,
        buckets=buckets)

    return oss_config


@contextmanager
def oss_progress(desc):
    progress = None

    def callback(i, n):
        nonlocal progress
        if progress is None:
            progress = tqdm(
                total=n,
                unit='B',
                unit_scale=True,
                unit_divisor=1024,
                leave=False,
                desc=desc,
                mininterval=1.0,
                maxinterval=5.0)
        progress.update(i - progress.n)

    yield callback
    if progress is not None:
        progress.close()


@contextmanager
def ignore_oss_error(msg=''):
    import oss2
    try:
        yield
    except (oss2.exceptions.RequestError, oss2.exceptions.ServerError) as e:
        logging.error(str(e) + ' ' + msg)


@contextmanager
def mute_stderr():
    cache = sys.stderr
    sys.stderr = StringIO()
    try:
        yield None
    finally:
        sys.stderr = cache



# Copyright (c) Alibaba, Inc. and its affiliates.
import random

import torch
from mmcv.runner import get_dist_info
from mmcv.runner.hooks import Hook
from torch import distributed as dist

from .registry import HOOKS


@HOOKS.register_module()
class SyncRandomSizeHook(Hook):
    """Change and synchronize the random image size across ranks, currently
    used in YOLOX.

    Args:
        ratio_range (tuple[int]): Random ratio range. It will be multiplied
            by 32, and then change the dataset output image size.
            Default: (14, 26).
        img_scale (tuple[int]): Size of input image. Default: (640, 640).
        interval (int): The interval of change image size. Default: 10.
        device (torch.device | str): device for returned tensors.
            Default: 'cuda'.
    """

    # TODO need to fix some bugs, to update by 10 iters but when training update by 1 epoch
    def __init__(
            self,
            ratio_range=(14, 26),
            img_scale=(640, 640),
            interval=10,  # by iterations
            device='cuda',
            **kwargs):
        self.rank, world_size = get_dist_info()
        self.is_distributed = world_size > 1
        self.ratio_range = ratio_range
        self.img_scale = img_scale
        self.interval = interval
        self.device = device

    def after_train_iter(self, runner):
        """Change the dataset output image size."""
        if self.ratio_range is not None and (runner.iter +
                                             1) % self.interval == 0:
            # Due to DDP and DP get the device behavior inconsistent,
            # so we did not get the device from runner.model.
            tensor = torch.LongTensor(2).to(self.device)

            if self.rank == 0:
                size_factor = self.img_scale[1] * 1. / self.img_scale[0]
                size = random.randint(*self.ratio_range)
                size = (int(32 * size), 32 * int(size * size_factor))
                tensor[0] = size[0]
                tensor[1] = size[1]

            if self.is_distributed:
                dist.barrier()
                dist.broadcast(tensor, 0)

            # TODO some bugs need fix, to update by 10 iters but when training update by 1 epoch
            runner.data_loader.dataset.update_dynamic_scale(
                (tensor[0].item(), tensor[1].item()))



# Copyright (c) Alibaba, Inc. and its affiliates.
import time
from collections.abc import Sequence

import numpy as np

from easycv.datasets.registry import PIPELINES
from easycv.file.image import load_image
from easycv.framework.errors import TypeError
from easycv.utils.registry import build_from_cfg


@PIPELINES.register_module()
class Compose(object):
    """Compose a data pipeline with a sequence of transforms.
    Args:
        transforms (list[dict | callable]):
            Either config dicts of transforms or transform objects.
    """

    def __init__(self, transforms, profiling=False):
        assert isinstance(transforms, Sequence)
        self.profiling = profiling
        self.transforms = []
        for transform in transforms:
            if isinstance(transform, dict):
                transform = build_from_cfg(transform, PIPELINES)
                self.transforms.append(transform)
            elif callable(transform):
                self.transforms.append(transform)
            else:
                raise TypeError('transform must be callable or a dict, but got'
                                f' {type(transform)}')

    def __call__(self, data):
        for t in self.transforms:
            if self.profiling:
                start = time.time()

            data = t(data)

            if self.profiling:
                print(f'{t} time {time.time()-start}')

            if data is None:
                return None
        return data

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += f'\n    {t}'
        format_string += '\n)'
        return format_string


@PIPELINES.register_module()
class LoadImage:
    """Load an image from file or numpy or PIL object.
    Args:
        to_float32 (bool): Whether to convert the loaded image to a float32
            numpy array. If set to False, the loaded image is an uint8 array.
            Defaults to False.
    """

    def __init__(self, to_float32=False, mode='bgr'):
        self.to_float32 = to_float32
        self.mode = mode

    def __call__(self, results):
        """Call functions to load image and get image meta information.
        Returns:
            dict: The dict contains loaded image and meta information.
        """
        filename = results.get('filename', None)
        img = results.get('img', None)

        if img is not None:
            if not isinstance(img, np.ndarray):
                img = np.asarray(img, dtype=np.uint8)
        else:
            assert filename is not None, 'Please provide "filename" or "img"!'
            img = load_image(filename, mode=self.mode)

        if self.to_float32:
            img = img.astype(np.float32)

        results['filename'] = filename
        results['img'] = img
        results['img_shape'] = img.shape
        results['ori_shape'] = img.shape
        results['ori_img_shape'] = img.shape
        results['img_fields'] = ['img']
        return results

    def __repr__(self):
        repr_str = (f'{self.__class__.__name__}('
                    f'to_float32={self.to_float32}, '
                    f"mode='{self.mode}'")

        return repr_str



import copy
import math

import numpy as np
import torch
import torch.nn as nn

from easycv.models.builder import LOSSES

CONSTANT_CONTOUR = 66
CONSTANT_EYEBROW = 18
CONSTANT_EYE = 18
CONSTANT_NOSE = 30
CONSTANT_LIPS = 40
CONSTANT_EYE_CENTER = 4


@LOSSES.register_module()
class WingLossWithPose(nn.Module):

    def __init__(self,
                 num_points=106,
                 left_eye_left_corner_index=66,
                 right_eye_right_corner_index=79,
                 points_weight=1.0,
                 contour_weight=1.5,
                 eyebrow_weight=1.5,
                 eye_weight=1.7,
                 nose_weight=1.3,
                 lip_weight=1.7,
                 omega=10,
                 epsilon=2):
        super(WingLossWithPose, self).__init__()
        self.omega = omega
        self.epsilon = epsilon

        self.num_points = num_points
        self.left_eye_left_corner_index = left_eye_left_corner_index
        self.right_eye_right_corner_index = right_eye_right_corner_index
        self.points_weight = points_weight
        contour_weight = np.full(CONSTANT_CONTOUR, contour_weight)
        eyebrow_left_weight = np.full(CONSTANT_EYEBROW, eyebrow_weight)
        eyebrow_right_weight = np.full(CONSTANT_EYEBROW, eyebrow_weight)
        nose_weight = np.full(CONSTANT_NOSE, nose_weight)
        eye_left_weight = np.full(CONSTANT_EYE, eye_weight)
        eye_right_weight = np.full(CONSTANT_EYE, eye_weight)
        lips_weight = np.full(CONSTANT_LIPS, lip_weight)
        eye_center_weight = np.full(CONSTANT_EYE_CENTER, eye_weight)
        part_weight = np.concatenate(
            (contour_weight, eyebrow_left_weight, eyebrow_right_weight,
             nose_weight, eye_left_weight, eye_right_weight, lips_weight,
             eye_center_weight),
            axis=0)

        self.part_weight = None
        if part_weight is not None:
            self.part_weight = torch.from_numpy(part_weight)

    def forward(self, pred, target, pose):
        weight = 5.0 * (1.0 - torch.cos(pose * np.pi / 180.0)) + 1.0
        weight = torch.sum(weight, dim=1) / 3.0
        weight = weight.view((weight.shape[0], 1))
        self.part_weight = self.part_weight.to(weight.device)
        if self.part_weight is not None:
            weight = weight * self.part_weight

        y = target
        y_hat = pred
        delta_y = (y - y_hat).abs() * weight
        delta_y1 = delta_y[delta_y < self.omega]
        delta_y2 = delta_y[delta_y >= self.omega]
        loss1 = self.omega * torch.log(1 + delta_y1 / self.epsilon)
        C = self.omega - self.omega * math.log(1 + self.omega / self.epsilon)
        loss = delta_y2 - C
        result = self.points_weight * (loss1.sum() + loss.sum()) / (
            len(loss1) + len(loss))

        return result


@LOSSES.register_module()
class FacePoseLoss(nn.Module):

    def __init__(self, pose_weight=1.0):
        super(FacePoseLoss, self).__init__()
        self.criterion = nn.MSELoss()
        self.pose_weight = pose_weight

    def forward(self, pred, target):
        result = self.pose_weight * self.criterion(pred, target)
        return result



_base_ = './segformer_b0_coco.py'

CLASSES = [
    'person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus', 'train',
    'truck', 'boat', 'traffic light', 'fire hydrant', 'stop sign',
    'parking meter', 'bench', 'bird', 'cat', 'dog', 'horse', 'sheep', 'cow',
    'elephant', 'bear', 'zebra', 'giraffe', 'backpack', 'umbrella', 'handbag',
    'tie', 'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball', 'kite',
    'baseball bat', 'baseball glove', 'skateboard', 'surfboard',
    'tennis racket', 'bottle', 'wine glass', 'cup', 'fork', 'knife', 'spoon',
    'bowl', 'banana', 'apple', 'sandwich', 'orange', 'broccoli', 'carrot',
    'hot dog', 'pizza', 'donut', 'cake', 'chair', 'couch', 'potted plant',
    'bed', 'dining table', 'toilet', 'tv', 'laptop', 'mouse', 'remote',
    'keyboard', 'cell phone', 'microwave', 'oven', 'toaster', 'sink',
    'refrigerator', 'book', 'clock', 'vase', 'scissors', 'teddy bear',
    'hair drier', 'toothbrush', 'banner', 'blanket', 'branch', 'bridge',
    'building-other', 'bush', 'cabinet', 'cage', 'cardboard', 'carpet',
    'ceiling-other', 'ceiling-tile', 'cloth', 'clothes', 'clouds', 'counter',
    'cupboard', 'curtain', 'desk-stuff', 'dirt', 'door-stuff', 'fence',
    'floor-marble', 'floor-other', 'floor-stone', 'floor-tile', 'floor-wood',
    'flower', 'fog', 'food-other', 'fruit', 'furniture-other', 'grass',
    'gravel', 'ground-other', 'hill', 'house', 'leaves', 'light', 'mat',
    'metal', 'mirror-stuff', 'moss', 'mountain', 'mud', 'napkin', 'net',
    'paper', 'pavement', 'pillow', 'plant-other', 'plastic', 'platform',
    'playingfield', 'railing', 'railroad', 'river', 'road', 'rock', 'roof',
    'rug', 'salad', 'sand', 'sea', 'shelf', 'sky-other', 'skyscraper', 'snow',
    'solid-other', 'stairs', 'stone', 'straw', 'structural-other', 'table',
    'tent', 'textile-other', 'towel', 'tree', 'vegetable', 'wall-brick',
    'wall-concrete', 'wall-other', 'wall-panel', 'wall-stone', 'wall-tile',
    'wall-wood', 'water-other', 'waterdrops', 'window-blind', 'window-other',
    'wood'
]

model = dict(
    pretrained=
    'https://download.openmmlab.com/mmsegmentation/v0.5/pretrain/segformer/mit_b5_20220624-658746d9.pth',
    backbone=dict(
        embed_dims=64,
        num_layers=[3, 6, 40, 3],
    ),
    decode_head=dict(
        in_channels=[64, 128, 320, 512],
        channels=768,
    ),
)

img_norm_cfg = dict(
    mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)

img_scale = (2048, 640)
crop_size = (640, 640)
train_pipeline = [
    dict(type='MMResize', img_scale=img_scale, ratio_range=(0.5, 2.0)),
    dict(type='SegRandomCrop', crop_size=crop_size, cat_max_ratio=0.75),
    dict(type='MMRandomFlip', flip_ratio=0.5),
    dict(type='MMPhotoMetricDistortion'),
    dict(type='MMNormalize', **img_norm_cfg),
    dict(type='MMPad', size=crop_size, pad_val=dict(img=0, masks=0, seg=255)),
    dict(type='DefaultFormatBundle'),
    dict(
        type='Collect',
        keys=['img', 'gt_semantic_seg'],
        meta_keys=('filename', 'ori_filename', 'ori_shape', 'img_shape',
                   'pad_shape', 'scale_factor', 'flip', 'flip_direction',
                   'img_norm_cfg')),
]
test_pipeline = [
    dict(
        type='MMMultiScaleFlipAug',
        img_scale=img_scale,
        # img_ratios=[0.5, 0.75, 1.0, 1.25, 1.5, 1.75],
        flip=False,
        transforms=[
            dict(type='MMResize', keep_ratio=True),
            dict(type='MMRandomFlip'),
            dict(type='MMNormalize', **img_norm_cfg),
            dict(type='ImageToTensor', keys=['img']),
            dict(
                type='Collect',
                keys=['img'],
                meta_keys=('filename', 'ori_filename', 'ori_shape',
                           'img_shape', 'pad_shape', 'scale_factor', 'flip',
                           'flip_direction', 'img_norm_cfg')),
        ])
]

data_root = './data/coco_stuff164k/'
# dataset settings
data_type = 'SegSourceRaw'
data_root = 'data/VOCdevkit/VOC2012'

train_img_root = data_root + 'JPEGImages'
train_label_root = data_root + 'SegmentationClass'
train_list_file = data_root + 'ImageSets/Segmentation/train.txt'

val_img_root = data_root + 'JPEGImages'
val_label_root = data_root + 'SegmentationClass'
val_list_file = data_root + 'ImageSets/Segmentation/val.txt'

test_batch_size = 2

data = dict(
    imgs_per_gpu=2,
    workers_per_gpu=2,
    train=dict(
        type='SegDataset',
        ignore_index=255,
        data_source=dict(
            type=data_type,
            img_suffix='.jpg',
            label_suffix='.png',
            img_root=train_img_root,
            label_root=train_label_root,
            split=train_list_file,
            classes=CLASSES,
        ),
        pipeline=train_pipeline),
    val=dict(
        imgs_per_gpu=test_batch_size,
        ignore_index=255,
        type='SegDataset',
        data_source=dict(
            type=data_type,
            img_suffix='.jpg',
            label_suffix='.png',
            img_root=val_img_root,
            label_root=val_label_root,
            split=val_list_file,
            classes=CLASSES,
        ),
        pipeline=test_pipeline),
)



# Copyright (c) Alibaba, Inc. and its affiliates.
import warnings

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import ConvModule

from easycv.models.builder import HEADS
from easycv.models.utils.ops import resize_tensor
from .base import BaseDecodeHead


# Modified from https://github.com/open-mmlab/mmsegmentation/blob/master/mmseg/models/decode_heads/uper_head.py
@HEADS.register_module()
class UPerHead(BaseDecodeHead):
    """Unified Perceptual Parsing for Scene Understanding.

    This head is the implementation of `UPerNet
    <https://arxiv.org/abs/1807.10221>`_.

    Args:
        pool_scales (tuple[int]): Pooling scales used in Pooling Pyramid
            Module applied on the last feature. Default: (1, 2, 3, 6).
    """

    def __init__(self, pool_scales=(1, 2, 3, 6), **kwargs):
        super(UPerHead, self).__init__(
            input_transform='multiple_select', **kwargs)
        # PSP Module
        self.psp_modules = PPM(
            pool_scales,
            self.in_channels[-1],
            self.channels,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
            align_corners=self.align_corners)
        self.bottleneck = ConvModule(
            self.in_channels[-1] + len(pool_scales) * self.channels,
            self.channels,
            3,
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg)
        # FPN Module
        self.lateral_convs = nn.ModuleList()
        self.fpn_convs = nn.ModuleList()
        for in_channels in self.in_channels[:-1]:  # skip the top layer
            l_conv = ConvModule(
                in_channels,
                self.channels,
                1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
                inplace=False)
            fpn_conv = ConvModule(
                self.channels,
                self.channels,
                3,
                padding=1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
                inplace=False)
            self.lateral_convs.append(l_conv)
            self.fpn_convs.append(fpn_conv)

        self.fpn_bottleneck = ConvModule(
            len(self.in_channels) * self.channels,
            self.channels,
            3,
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg)

    def psp_forward(self, inputs):
        """Forward function of PSP module."""
        x = inputs[-1]
        psp_outs = [x]
        psp_outs.extend(self.psp_modules(x))
        psp_outs = torch.cat(psp_outs, dim=1)
        output = self.bottleneck(psp_outs)

        return output

    def _forward_feature(self, inputs):
        """Forward function for feature maps before classifying each pixel with
        ``self.cls_seg`` fc.

        Args:
            inputs (list[Tensor]): List of multi-level img features.

        Returns:
            feats (Tensor): A tensor of shape (batch_size, self.channels,
                H, W) which is feature map for last layer of decoder head.
        """
        inputs = self._transform_inputs(inputs)

        # build laterals
        laterals = [
            lateral_conv(inputs[i])
            for i, lateral_conv in enumerate(self.lateral_convs)
        ]

        laterals.append(self.psp_forward(inputs))

        # build top-down path
        used_backbone_levels = len(laterals)
        for i in range(used_backbone_levels - 1, 0, -1):
            prev_shape = laterals[i - 1].shape[2:]
            laterals[i - 1] = laterals[i - 1] + resize_tensor(
                laterals[i],
                size=prev_shape,
                mode='bilinear',
                align_corners=self.align_corners)

        # build outputs
        fpn_outs = [
            self.fpn_convs[i](laterals[i])
            for i in range(used_backbone_levels - 1)
        ]
        # append psp feature
        fpn_outs.append(laterals[-1])

        for i in range(used_backbone_levels - 1, 0, -1):
            fpn_outs[i] = resize_tensor(
                fpn_outs[i],
                size=fpn_outs[0].shape[2:],
                mode='bilinear',
                align_corners=self.align_corners)
        fpn_outs = torch.cat(fpn_outs, dim=1)
        feats = self.fpn_bottleneck(fpn_outs)
        return feats

    def forward(self, inputs):
        """Forward function."""
        output = self._forward_feature(inputs)
        output = self.cls_seg(output)
        return output


class PPM(nn.ModuleList):
    """Pooling Pyramid Module used in PSPNet.

    Args:
        pool_scales (tuple[int]): Pooling scales used in Pooling Pyramid
            Module.
        in_channels (int): Input channels.
        channels (int): Channels after modules, before conv_seg.
        conv_cfg (dict|None): Config of conv layers.
        norm_cfg (dict|None): Config of norm layers.
        act_cfg (dict): Config of activation layers.
        align_corners (bool): align_corners argument of F.interpolate.
    """

    def __init__(self, pool_scales, in_channels, channels, conv_cfg, norm_cfg,
                 act_cfg, align_corners, **kwargs):
        super(PPM, self).__init__()
        self.pool_scales = pool_scales
        self.align_corners = align_corners
        self.in_channels = in_channels
        self.channels = channels
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        for pool_scale in pool_scales:
            self.append(
                nn.Sequential(
                    nn.AdaptiveAvgPool2d(pool_scale),
                    ConvModule(
                        self.in_channels,
                        self.channels,
                        1,
                        conv_cfg=self.conv_cfg,
                        norm_cfg=self.norm_cfg,
                        act_cfg=self.act_cfg,
                        **kwargs)))

    def forward(self, x):
        """Forward function."""
        ppm_outs = []
        for ppm in self:
            ppm_out = ppm(x)
            upsampled_ppm_out = resize_tensor(
                ppm_out,
                size=x.size()[2:],
                mode='bilinear',
                align_corners=self.align_corners)
            ppm_outs.append(upsampled_ppm_out)
        return ppm_outs



_base_ = './yolox_s_8xb16_300e_coco.py'

# s m l x
img_scale = (640, 640)
random_size = (14, 26)
scale_ratio = (0.1, 2)

# tiny nano without mixup
# img_scale = (416, 416)
# random_size = (10, 20)
# scale_ratio = (0.5, 1.5)

CLASSES = [
    'aeroplane', 'bicycle', 'bird', 'boat', 'bottle', 'bus', 'car', 'cat',
    'chair', 'cow', 'diningtable', 'dog', 'horse', 'motorbike', 'person',
    'pottedplant', 'sheep', 'sofa', 'train', 'tvmonitor'
]

model = dict(num_classes=20)

# dataset settings
data_root = 'data/voc/'
img_norm_cfg = dict(
    mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)

train_pipeline = [
    dict(type='MMMosaic', img_scale=img_scale, pad_val=114.0),
    dict(
        type='MMRandomAffine',
        scaling_ratio_range=scale_ratio,
        border=(-img_scale[0] // 2, -img_scale[1] // 2)),
    dict(
        type='MMMixUp',  # s m x l; tiny nano will detele
        img_scale=img_scale,
        ratio_range=(0.8, 1.6),
        pad_val=114.0),
    dict(
        type='MMPhotoMetricDistortion',
        brightness_delta=32,
        contrast_range=(0.5, 1.5),
        saturation_range=(0.5, 1.5),
        hue_delta=18),
    dict(type='MMRandomFlip', flip_ratio=0.5),
    dict(type='MMResize', keep_ratio=True),
    dict(type='MMPad', pad_to_square=True, pad_val=(114.0, 114.0, 114.0)),
    dict(type='MMNormalize', **img_norm_cfg),
    dict(type='DefaultFormatBundle'),
    dict(type='Collect', keys=['img', 'gt_bboxes', 'gt_labels'])
]
test_pipeline = [
    dict(type='MMResize', img_scale=img_scale, keep_ratio=True),
    dict(type='MMPad', pad_to_square=True, pad_val=(114.0, 114.0, 114.0)),
    dict(type='MMNormalize', **img_norm_cfg),
    dict(type='DefaultFormatBundle'),
    dict(type='Collect', keys=['img'])
]

train_dataset = dict(
    type='DetImagesMixDataset',
    data_source=dict(
        type='DetSourceVOC',
        path=data_root + 'ImageSets/Main/train.txt',
        classes=CLASSES,
        cache_at_init=True),
    pipeline=train_pipeline,
    dynamic_scale=img_scale)

val_dataset = dict(
    type='DetImagesMixDataset',
    imgs_per_gpu=2,
    data_source=dict(
        type='DetSourceVOC',
        path=data_root + 'ImageSets/Main/val.txt',
        classes=CLASSES,
        cache_at_init=True),
    pipeline=test_pipeline,
    dynamic_scale=None,
    label_padding=False)

data = dict(
    imgs_per_gpu=16, workers_per_gpu=4, train=train_dataset, val=val_dataset)

# # evaluation
eval_pipelines = [
    dict(
        mode='test',
        data=data['val'],
        evaluators=[dict(type='CocoDetectionEvaluator', classes=CLASSES)],
    )
]



# Copyright (c) Alibaba, Inc. and its affiliates.
from . import data_sources  # pylint: disable=unused-import
from . import pipelines  # pylint: disable=unused-import
from .hand_coco_wholebody_dataset import HandCocoWholeBodyDataset
from .top_down import PoseTopDownDataset
from .wholebody_topdown_coco_dataset import WholeBodyCocoTopDownDataset

__all__ = [
    'PoseTopDownDataset', 'HandCocoWholeBodyDataset',
    'WholeBodyCocoTopDownDataset'
]



# Copyright (c) Alibaba, Inc. and its affiliates.
"""
isort:skip_file
"""
from __future__ import division
import argparse
import importlib
import json
import os
import os.path as osp
import sys
sys.path.append(os.path.abspath(os.path.dirname(os.path.dirname(__file__))))
sys.path.append(
    os.path.abspath(
        osp.join(os.path.dirname(os.path.dirname(__file__)), '../')))

# adapt to torchacc, init before some torch imports
from easycv.utils.torchacc_util import is_torchacc_enabled
if is_torchacc_enabled():
    from easycv.toolkit.torchacc import torchacc_init
    torchacc_init()

import time
import requests
import torch
import torch.distributed as dist
from mmcv.runner import init_dist
from mmcv import DictAction

from easycv import __version__
from easycv.apis import init_random_seed, set_random_seed, train_model
from easycv.datasets import build_dataloader, build_dataset
from easycv.datasets.utils import is_dali_dataset_type
from easycv.file import io
from easycv.models import build_model
from easycv.utils.collect_env import collect_env
from easycv.utils.logger import get_root_logger
from easycv.utils import mmlab_utils
from easycv.utils.config_tools import (traverse_replace, CONFIG_TEMPLATE_ZOO,
                                       mmcv_config_fromfile,
                                       pai_config_fromfile)
from easycv.utils.dist_utils import get_device, is_master
from easycv.utils.setup_env import setup_multi_processes


def parse_args():
    parser = argparse.ArgumentParser(description='Train a model')
    parser.add_argument(
        'config', help='train config file path', type=str, default=None)
    parser.add_argument(
        '--work_dir',
        type=str,
        default=None,
        help='the dir to save logs and models')
    parser.add_argument(
        '--resume_from', help='the checkpoint file to resume from')
    parser.add_argument('--load_from', help='the checkpoint file to load from')
    parser.add_argument(
        '--pretrained', default=None, help='pretrained model file')
    parser.add_argument(
        '--no_validate',
        action='store_true',
        help='whether not to evaluate the checkpoint during training')
    parser.add_argument(
        '--gpus',
        type=int,
        default=1,
        help='number of gpus to use '
        '(only applicable to non-distributed training)')
    parser.add_argument('--seed', type=int, default=None, help='random seed')
    parser.add_argument(
        '--diff-seed',
        action='store_true',
        help='Whether or not set different seeds for different ranks')
    parser.add_argument('--fp16', action='store_true', help='use fp16')
    parser.add_argument(
        '--deterministic',
        action='store_true',
        help='whether to set deterministic options for CUDNN backend.')
    parser.add_argument(
        '--launcher',
        choices=['none', 'pytorch', 'slurm', 'mpi'],
        default='none',
        help='job launcher')
    parser.add_argument('--local_rank', type=int, default=0)
    parser.add_argument(
        '--port',
        type=int,
        default=29500,
        help='port only works when launcher=="slurm"')
    parser.add_argument(
        '--model_type',
        type=str,
        default=None,
        help=
        'parameterize param when user specific choose a model config template like CLASSIFICATION: classification.py'
    )
    parser.add_argument(
        '--user_config_params',
        nargs='+',
        action=DictAction,
        help='override some settings in the used config, the key-value pair '
        'in xxx=yyy format will be merged into config file. If the value to '
        'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
        'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
        'Note that the quotation marks are necessary and that no white space '
        'is allowed. Single quote double quote equivalent.')

    args = parser.parse_args()
    if 'LOCAL_RANK' not in os.environ:
        os.environ['LOCAL_RANK'] = str(args.local_rank)

    return args


def main():
    args = parse_args()

    if args.model_type is not None:
        assert args.model_type in CONFIG_TEMPLATE_ZOO, 'model_type must be in [%s]' % (
            ', '.join(CONFIG_TEMPLATE_ZOO.keys()))
        print('model_type=%s, config file will be replaced by %s' %
              (args.model_type, CONFIG_TEMPLATE_ZOO[args.model_type]))
        args.config = CONFIG_TEMPLATE_ZOO[args.model_type]

    if args.config.startswith('http'):

        r = requests.get(args.config)
        # download config in current dir
        tpath = args.config.split('/')[-1]
        while not osp.exists(tpath):
            try:
                with open(tpath, 'wb') as code:
                    code.write(r.content)
            except:
                pass

        args.config = tpath

    # build cfg
    if args.user_config_params is None:
        cfg = mmcv_config_fromfile(args.config)
    else:
        cfg = pai_config_fromfile(args.config, args.user_config_params,
                                  args.model_type)

    # set multi-process settings
    setup_multi_processes(cfg)

    # set cudnn_benchmark
    if cfg.get('cudnn_benchmark', False):
        torch.backends.cudnn.benchmark = True

    # update configs according to CLI args
    if args.work_dir is not None:
        cfg.work_dir = args.work_dir

    if cfg.get('work_dir', None) is None:
        cfg.work_dir = './work_dir'

    # if `work_dir` is oss path, redirect `work_dir` to local path, add `oss_work_dir` point to oss path,
    # and use osssync hook to upload log and ckpt in work_dir to oss_work_dir
    if cfg.work_dir.startswith('oss://'):
        cfg.oss_work_dir = cfg.work_dir
        cfg.work_dir = osp.join('work_dirs',
                                cfg.work_dir.replace('oss://', ''))
    else:
        cfg.oss_work_dir = None

    if args.resume_from is not None and len(args.resume_from) > 0:
        cfg.resume_from = args.resume_from
    if args.load_from is not None and len(args.load_from) > 0:
        cfg.load_from = args.load_from

    # dynamic adapt mmdet models
    mmlab_utils.dynamic_adapt_for_mmlab(cfg)

    cfg.gpus = args.gpus

    # check memcached package exists
    if importlib.util.find_spec('mc') is None:
        traverse_replace(cfg, 'memcached', False)

    # check oss_config and init oss io
    if cfg.get('oss_io_config', None) is not None:
        io.access_oss(**cfg.oss_io_config)
    # init distributed env first, since logger depends on the dist info.
    if not is_torchacc_enabled():
        if args.launcher == 'none':
            assert cfg.model.type not in \
                ['DeepCluster', 'MOCO', 'SimCLR', 'ODC', 'NPID'], \
                '{} does not support non-dist training.'.format(cfg.model.type)
        else:
            if args.launcher == 'slurm':
                cfg.dist_params['port'] = args.port
            init_dist(args.launcher, **cfg.dist_params)

    distributed = torch.cuda.is_available(
    ) and torch.distributed.is_initialized()

    # create work_dir
    if not io.exists(cfg.work_dir):
        io.makedirs(cfg.work_dir)
    # init the logger before other steps
    timestamp = time.strftime('%Y%m%d_%H%M%S', time.localtime())
    log_file = osp.join(cfg.work_dir, '{}.log'.format(timestamp))
    logger = get_root_logger(log_file=log_file, log_level=cfg.log_level)

    # init the meta dict to record some important information such as
    # environment info and seed, which will be logged
    meta = dict()
    # log env info
    env_info_dict = collect_env()
    env_info = '\n'.join([('{}: {}'.format(k, v))
                          for k, v in env_info_dict.items()])
    dash_line = '-' * 60 + '\n'
    logger.info('Environment info:\n' + dash_line + env_info + '\n' +
                dash_line)
    meta['env_info'] = env_info

    # log some basic info
    logger.info('Distributed training: {}'.format(distributed))
    logger.info('Config:\n{}'.format(cfg.text))
    logger.info('Config Dict:\n{}'.format(json.dumps(cfg._cfg_dict)))
    logger.info('GPU INFO : {}'.format(torch.cuda.get_device_name(0)))

    # set random seeds
    # Using different seeds for different ranks may reduce accuracy
    seed = init_random_seed(args.seed, device=get_device())
    seed = seed + dist.get_rank() if args.diff_seed else seed
    if is_torchacc_enabled():
        assert seed is not None, 'Must provide `seed` to sync model initializer if use torchacc!'

    if seed is not None:
        logger.info('Set random seed to {}, deterministic: {}'.format(
            seed, args.deterministic))
        set_random_seed(seed, deterministic=args.deterministic)
    cfg.seed = seed
    meta['seed'] = seed

    if args.pretrained is not None:
        assert isinstance(args.pretrained, str)
        cfg.model.pretrained = args.pretrained
    model = build_model(cfg.model)

    if is_master():
        print(model)

    if 'stage' in cfg.model and cfg.model['stage'] == 'EDGE':
        from easycv.utils.flops_counter import get_model_info
        get_model_info(model, cfg.img_scale, cfg.model, logger)

    assert len(cfg.workflow) == 1, 'Validation is called by hook.'
    if cfg.checkpoint_config is not None:
        # save easycv version, config file content and class names in
        # checkpoints as meta data
        cfg.checkpoint_config.meta = dict(
            easycv_version=__version__, config=cfg.text)

    # build dataloader
    if not is_dali_dataset_type(cfg.data.train['type']):
        shuffle = cfg.data.train.pop('shuffle', True)
        print(f'data shuffle: {shuffle}')

        # for odps data_source
        if cfg.data.train.data_source.type == 'OdpsReader' and cfg.data.train.data_source.get(
                'odps_io_config', None) is None:
            cfg.data.train.data_source['odps_io_config'] = cfg.get(
                'odps_io_config', None)
            assert (
                cfg.data.train.data_source.get('odps_io_config',
                                               None) is not None
            ), 'odps config must be set in cfg file / cfg.data.train.data_source !!'
            shuffle = False

        if getattr(cfg.data, 'pin_memory', False):
            mmlab_utils.fix_dc_pin_memory()
        datasets = [build_dataset(cfg.data.train)]
        data_loaders = [
            build_dataloader(
                ds,
                cfg.data.imgs_per_gpu,
                cfg.data.workers_per_gpu,
                cfg.gpus,
                dist=distributed,
                shuffle=shuffle,
                pin_memory=getattr(cfg.data, 'pin_memory', False),
                replace=getattr(cfg.data, 'sampling_replace', False),
                seed=cfg.seed,
                # The default should be set to True, because sometimes the last batch is not sampled enough, causing an error in batchnorm
                drop_last=getattr(cfg.data, 'drop_last', True),
                reuse_worker_cache=cfg.data.get('reuse_worker_cache', False),
                persistent_workers=cfg.data.get('persistent_workers', False),
                collate_hooks=cfg.data.get('train_collate_hooks', []),
                use_repeated_augment_sampler=cfg.data.get(
                    'use_repeated_augment_sampler', False)) for ds in datasets
        ]
    else:
        default_args = dict(
            batch_size=cfg.data.imgs_per_gpu,
            workers_per_gpu=cfg.data.workers_per_gpu,
            distributed=distributed)
        dataset = build_dataset(cfg.data.train, default_args)
        data_loaders = [dataset.get_dataloader()]

    # # add an attribute for visualization convenience
    train_model(
        model,
        data_loaders,
        cfg,
        distributed=distributed,
        timestamp=timestamp,
        meta=meta,
        validate=(not args.no_validate),
        use_fp16=args.fp16)


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
# Copyright (c) Alibaba, Inc. and its affiliates.
import glob
import os
import os.path as osp
import re
import tempfile
from argparse import ArgumentParser

import mmcv

from easycv.predictors import ClassificationPredictor, DetectionPredictor, ReIDPredictor
from easycv.thirdparty.mot.bytetrack.byte_tracker import BYTETracker
from easycv.thirdparty.mot.utils import (
    _is_valid_video, detection_result_filter, gen_res,
    get_mtmct_matching_results, prepare_crop_imgs, save_mtmct_crops,
    save_mtmct_vis_results, show_result, sub_cluster, trajectory_fusion,
    video2frames)


def main():
    parser = ArgumentParser()
    parser.add_argument('--det_config', default=None, help='detection config file')
    parser.add_argument('--det_checkpoint', help='detection checkpoint file')
    parser.add_argument('--reid_config', default=None, help='reid config file')
    parser.add_argument('--reid_checkpoint', help='reid checkpoint file')
    parser.add_argument('--input', help='input video file folder')
    parser.add_argument(
        '--output', help='output video file (mp4 format) or folder')
    parser.add_argument(
    '--use_folder',
    action='store_true',
    help='Enter the video images folder')
    args = parser.parse_args()
    assert args.output

    cid = 0
    cid_bias = dict()
    mot_list_breaks = []
    cid_tid_dict = dict()
    # load images
    seqs = os.listdir(args.input)
    for seq in sorted(seqs):
        fpath = os.path.join(args.input, seq)
        if args.use_folder:
            print('use image folder!')
            imgs = glob.glob(os.path.join(fpath, '*.jpg'))
            imgs.sort()
            assert len(imgs) > 0, '{} has no images.'.format(fpath)
        else:
            if os.path.isfile(fpath) and _is_valid_video(fpath):
                seq = seq.split('.')[-2]
                print('start tracking seq: {}'.format(seq))
                seq_dir = os.path.join(args.input, seq)
                if not os.path.exists(seq_dir):
                    print('ffmpeg processing of video {}'.format(fpath))
                    frames_path = video2frames(
                        video_path=fpath, outpath=args.input, frame_rate=25)
                else:
                    print('ffmpeg has been successfully run. The {} directory already exists'.format(seq_dir))

                fpath = seq_dir
                if os.path.isdir(fpath) == False:
                    print('{} is not a image folder.'.format(fpath))
                    continue
                assert os.path.isdir(fpath), '{} should be a directory'.format(
                    fpath)
                imgs = glob.glob(os.path.join(fpath, '*.jpg'))
                imgs.sort()
                assert len(imgs) > 0, '{} has no images.'.format(fpath)
            else:
                continue

        # build the model from a config file and a checkpoint file
        print('build det model and reid model!')
        det_model = DetectionPredictor(
            args.det_checkpoint, args.det_config, score_threshold=0.2) # detr-like score_threshold set to 0
        reid_model = ReIDPredictor(args.reid_checkpoint,
                                             args.reid_config, batch_size=256)
        tracker = BYTETracker(
            det_high_thresh=0.2,
            det_low_thresh=0.05,
            match_thresh=1.0,
            match_thresh_second=1.0,
            match_thresh_init=1.0,
            track_buffer=2,
            frame_rate=25)

        prog_bar = mmcv.ProgressBar(len(imgs))

        # test and show/save the images
        print('start tracking!')
        track_result = None
        mot_features_dict = dict()
        for frame_id, img in enumerate(imgs):
            result = det_model(img)[0]

            detection_boxes = result['detection_boxes']
            detection_scores = result['detection_scores']
            detection_classes = result['detection_classes']
            img_metas = result['img_metas']

            detection_boxes, detection_scores, detection_classes = detection_result_filter(
                detection_boxes,
                detection_scores,
                detection_classes,
                target_classes=[0], # 0: person 2: car
                target_thresholds=[0])
            if len(detection_boxes) > 0:
                track_result = tracker.update(
                    detection_boxes, detection_scores,
                    detection_classes)  # [id, t, l, b, r, score]

                batch_crop_imgs, track_bboxes = prepare_crop_imgs(
                    {
                        'boxes': track_result['track_bboxes'],
                        'img_metas': img_metas
                    })
                
                pred_embeddings = reid_model(batch_crop_imgs)['img_feature']

                for idx in range(len(track_bboxes)):
                    _id = int(track_bboxes[idx, 0])
                    imgname = f'{seq}_{_id}_{frame_id}.jpg'

                    mot_features_dict[imgname] = dict()
                    mot_features_dict[imgname]['bbox'] = track_bboxes[idx, 1:5]
                    mot_features_dict[imgname]['frame'] = f'{frame_id:06d}'
                    mot_features_dict[imgname]['id'] = _id
                    mot_features_dict[imgname]['imgname'] = imgname
                    mot_features_dict[imgname]['feat'] = pred_embeddings[idx].squeeze().numpy()
            
            prog_bar.update()

        cid_bias[cid] = float(0)
        tid_data, mot_list_break = trajectory_fusion(mot_features_dict, cid,
                                                     cid_bias)
        mot_list_breaks.append(mot_list_break)
        # single seq process
        for line in tid_data:
            tracklet = tid_data[line]
            tid = tracklet['tid']
            if (cid, tid) not in cid_tid_dict:
                cid_tid_dict[(cid, tid)] = tracklet
        
        cid += 1

    scene_cluster = list(cid_bias.keys())

    map_tid = sub_cluster(
        cid_tid_dict,
        scene_cluster,
        use_ff=False,
        use_rerank=False,
        use_st_filter=False)

    pred_mtmct_file = os.path.join(args.output, 'mtmct_result.txt')
    gen_res(pred_mtmct_file, scene_cluster, map_tid, mot_list_breaks)

    camera_results, cid_tid_fid_res = get_mtmct_matching_results(
        pred_mtmct_file, secs_interval=0.5, video_fps=20)

    crops_dir = os.path.join(args.output, 'mtmct_crops')
    save_mtmct_crops(
        cid_tid_fid_res, images_dir=args.input, crops_dir=crops_dir)

    save_dir = os.path.join(args.output, 'mtmct_vis')
    save_mtmct_vis_results(
        camera_results,
        images_dir=args.input,
        save_dir=save_dir,
        save_videos=True)


if __name__ == '__main__':
    main()


# Copyright (c) Alibaba, Inc. and its affiliates.
"""
isort:skip_file
"""
import time
import json
import argparse
import os
import os.path as osp
import sys
sys.path.append(os.path.abspath(os.path.dirname(os.path.dirname(__file__))))
sys.path.append(
    os.path.abspath(
        osp.join(os.path.dirname(os.path.dirname(__file__)), '../')))

import mmcv
import requests
import torch
from mmcv import DictAction
from mmcv.parallel import MMDataParallel, MMDistributedDataParallel
from mmcv.runner import get_dist_info, init_dist

from easycv import datasets
from easycv.apis import multi_gpu_test, single_gpu_test
from easycv.core.evaluation.builder import build_evaluator
from easycv.datasets import build_dataloader, build_dataset
from easycv.file import io
from easycv.models import build_model
from easycv.utils.checkpoint import load_checkpoint
from easycv.utils.config_tools import (CONFIG_TEMPLATE_ZOO,
                                       mmcv_config_fromfile, rebuild_config)
from easycv.utils.mmlab_utils import dynamic_adapt_for_mmlab

from easycv.utils.setup_env import setup_multi_processes
from easycv.framework.errors import ValueError, NotImplementedError
from easycv.utils.misc import reparameterize_models


def parse_args():
    parser = argparse.ArgumentParser(
        description='EasyCV test (and eval) a model')
    parser.add_argument('config', help='test config file path')
    parser.add_argument('checkpoint', help='checkpoint file')
    parser.add_argument(
        '--work_dir', help='the directory to save evaluation logs')
    parser.add_argument('--out', help='output result file in pickle format')
    # parser.add_argument(
    #     '--fuse-conv-bn',
    #     action='store_true',
    #     help='Whether to fuse conv and bn, this will slightly increase'
    #     'the inference speed')
    parser.add_argument(
        '--inference-only',
        action='store_true',
        help='save the output results without perform evaluation. It is'
        'useful when you want to format the result to a specific format and '
        'submit it to the test server')
    parser.add_argument('--eval', action='store_true', help='evaluate result')
    parser.add_argument('--fp16', action='store_true', help='use fp16')
    parser.add_argument('--show', action='store_true', help='show results')
    parser.add_argument(
        '--show-dir', help='directory where painted images will be saved')
    # parser.add_argument(
    #     '--show-score-thr',
    #     type=float,
    #     default=0.3,
    #     help='score threshold (default: 0.3)')
    parser.add_argument(
        '--gpu-collect',
        action='store_true',
        help='whether to use gpu to collect results.')
    parser.add_argument(
        '--tmpdir',
        help='tmp directory used for collecting results from multiple '
        'workers, available when gpu-collect is not specified')
    parser.add_argument(
        '--options', nargs='+', action=DictAction, help='arguments in dict')
    parser.add_argument(
        '--launcher',
        choices=['none', 'pytorch', 'slurm', 'mpi'],
        default='none',
        help='job launcher')
    parser.add_argument('--local_rank', type=int, default=0)
    parser.add_argument(
        '--model_type',
        type=str,
        default=None,
        help=
        'parameterize param when user specific choose a model config template like CLASSIFICATION: classification.py'
    )
    parser.add_argument(
        '--user_config_params',
        nargs=argparse.REMAINDER,
        default=None,
        help='modify config options using the command-line')
    args = parser.parse_args()
    if 'LOCAL_RANK' not in os.environ:
        os.environ['LOCAL_RANK'] = str(args.local_rank)
    return args


def main():
    args = parse_args()

    assert args.out or args.eval or args.inference_only or args.show \
        or args.show_dir, \
        ('Please specify at least one operation (save/eval/format/show the '
         'results / save the results) with the argument "--out", "--eval"'
         ', "--inference-only", "--show" or "--show-dir"')

    if args.eval and args.inference_only:
        raise ValueError(
            '--eval and --inference_only cannot be both specified')

    if args.out is not None and not args.out.endswith(('.pkl', '.pickle')):
        raise ValueError('The output file must be a pkl file.')

    if args.model_type is not None:
        assert args.model_type in CONFIG_TEMPLATE_ZOO, 'model_type must be in [%s]' % (
            ', '.join(CONFIG_TEMPLATE_ZOO.keys()))
        print('model_type=%s, config file will be replaced by %s' %
              (args.model_type, CONFIG_TEMPLATE_ZOO[args.model_type]))
        args.config = CONFIG_TEMPLATE_ZOO[args.model_type]

    if args.config.startswith('http'):

        r = requests.get(args.config)
        # download config in current dir
        tpath = args.config.split('/')[-1]
        while not osp.exists(tpath):
            try:
                with open(tpath, 'wb') as code:
                    code.write(r.content)
            except:
                pass

        args.config = tpath

    cfg = mmcv_config_fromfile(args.config)

    if args.user_config_params is not None:
        assert args.model_type is not None, 'model_type must be setted'
        # rebuild config by user config params
        cfg = rebuild_config(cfg, args.user_config_params)

    # check oss_config and init oss io
    if cfg.get('oss_io_config', None) is not None:
        io.access_oss(**cfg.oss_io_config)

    # set multi-process settings
    setup_multi_processes(cfg)

    # dynamic adapt mmdet models
    dynamic_adapt_for_mmlab(cfg)

    # set cudnn_benchmark
    if cfg.get('cudnn_benchmark', False):
        torch.backends.cudnn.benchmark = True
    cfg.model.pretrained = None
    if cfg.model.get('neck'):
        if type(cfg.model.neck) is list:
            pass
        else:
            if cfg.model.neck.get('rfp_backbone'):
                if cfg.model.neck.rfp_backbone.get('pretrained'):
                    cfg.model.neck.rfp_backbone.pretrained = None
    # cfg.data.test.test_mode = True

    # init distributed env first, since logger depends on the dist info.
    if args.launcher == 'none':
        distributed = False
    else:
        distributed = True
        init_dist(args.launcher, **cfg.dist_params)

    rank, _ = get_dist_info()

    if args.work_dir is not None and rank == 0:
        if not io.exists(args.work_dir):
            io.makedirs(args.work_dir)
        timestamp = time.strftime('%Y%m%d_%H%M%S', time.localtime())
        log_file = osp.join(args.work_dir, 'eval_{}.json'.format(timestamp))

    # build the model and load checkpoint
    model = build_model(cfg.model)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    print(f'use device {device}')
    checkpoint = load_checkpoint(model, args.checkpoint, map_location=device)

    # reparameter to deploy for RepVGG block
    model = reparameterize_models(model)

    model.to(device)
    # if args.fuse_conv_bn:
    #     model = fuse_module(model)

    # old versions did not save class info in checkpoints, this walkaround is
    # for backward compatibility
    if 'meta' in checkpoint and 'CLASSES' in checkpoint['meta']:
        model.CLASSES = checkpoint['meta']['CLASSES']
    elif hasattr(cfg, 'CLASSES'):
        model.CLASSES = cfg.CLASSES

    if not distributed:
        model = MMDataParallel(model, device_ids=[0])
    else:
        model = MMDistributedDataParallel(
            model.cuda(),
            device_ids=[torch.cuda.current_device()],
            broadcast_buffers=False)

    assert 'eval_pipelines' in cfg, 'eval_pipelines is needed for testting'
    for eval_pipe in cfg.eval_pipelines:
        eval_data = eval_pipe.get('data', None) or cfg.data.val
        # build the dataloader
        if eval_data.get('dali', False):
            data_loader = datasets.build_dali_dataset(
                eval_data).get_dataloader()
            # dali dataloader implements `evaluate` func, so use it as dummy dataset
            dataset = data_loader
        else:
            # dataset does not need imgs_per_gpu, except dali dataset
            imgs_per_gpu = eval_data.pop('imgs_per_gpu', cfg.data.imgs_per_gpu)

            dataset = build_dataset(eval_data)
            data_loader = build_dataloader(
                dataset,
                imgs_per_gpu=imgs_per_gpu,
                workers_per_gpu=cfg.data.workers_per_gpu,
                dist=distributed,
                shuffle=False)
            # oss_config=cfg.get('oss_io_config', None))

        if not distributed:
            outputs = single_gpu_test(
                model, data_loader, mode=eval_pipe.mode, use_fp16=args.fp16)
        else:
            outputs = multi_gpu_test(
                model,
                data_loader,
                mode=eval_pipe.mode,
                tmp_dir=args.tmpdir,
                gpu_collect=args.gpu_collect,
                use_fp16=args.fp16)

        if rank == 0:
            if args.out:
                print(f'\nwriting results to {args.out}')
                mmcv.dump(outputs, args.out)
            eval_kwargs = {}
            if args.options is not None:
                eval_kwargs.update(args.options)

            if args.inference_only:
                raise NotImplementedError('not implemented')
            if args.eval:
                for t in eval_pipe.evaluators:
                    if 'metric_type' in t:
                        t.pop('metric_type')
                evaluators = build_evaluator(eval_pipe.evaluators)
                eval_result = dataset.evaluate(outputs, evaluators=evaluators)
                print(f'\n eval_result {eval_result}')
                if args.work_dir is not None:
                    with io.open(log_file, 'w') as f:
                        json.dump(eval_result, f)


if __name__ == '__main__':
    main()



# oss_io_config = dict(
#     ak_id='your oss ak id',
#     ak_secret='your oss ak secret',
#     hosts='oss-cn-zhangjiakou.aliyuncs.com',  # your oss hosts
#     buckets=['your_bucket'])  # your oss buckets

oss_sync_config = dict(other_file_list=['**/events.out.tfevents*', '**/*log*'])

log_level = 'INFO'
load_from = None
resume_from = None
dist_params = dict(backend='nccl')
workflow = [('train', 1)]
checkpoint_config = dict(interval=10)

optimizer = dict(type='Adam', lr=5e-4)
optimizer_config = dict(grad_clip=None)
# learning policy
lr_config = dict(
    policy='step',
    warmup='linear',
    warmup_iters=500,
    warmup_ratio=0.001,
    step=[170, 200])
total_epochs = 210
log_config = dict(
    interval=50,
    hooks=[dict(type='TextLoggerHook'),
           dict(type='TensorboardLoggerHook')])
channel_cfg = dict(
    num_output_channels=21,
    dataset_joints=21,
    dataset_channel=[
        [
            0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18,
            19, 20
        ],
    ],
    inference_channel=[
        0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
        20
    ])

# model settings
model = dict(
    type='TopDown',
    pretrained=False,
    backbone=dict(
        type='HRNet',
        in_channels=3,
        extra=dict(
            stage1=dict(
                num_modules=1,
                num_branches=1,
                block='BOTTLENECK',
                num_blocks=(4, ),
                num_channels=(64, )),
            stage2=dict(
                num_modules=1,
                num_branches=2,
                block='BASIC',
                num_blocks=(4, 4),
                num_channels=(18, 36)),
            stage3=dict(
                num_modules=4,
                num_branches=3,
                block='BASIC',
                num_blocks=(4, 4, 4),
                num_channels=(18, 36, 72)),
            stage4=dict(
                num_modules=3,
                num_branches=4,
                block='BASIC',
                num_blocks=(4, 4, 4, 4),
                num_channels=(18, 36, 72, 144),
                multiscale_output=True),
            upsample=dict(mode='bilinear', align_corners=False))),
    keypoint_head=dict(
        type='TopdownHeatmapSimpleHead',
        in_channels=[18, 36, 72, 144],
        in_index=(0, 1, 2, 3),
        input_transform='resize_concat',
        out_channels=channel_cfg['num_output_channels'],
        num_deconv_layers=0,
        extra=dict(
            final_conv_kernel=1, num_conv_layers=1, num_conv_kernels=(1, )),
        loss_keypoint=dict(type='JointsMSELoss', use_target_weight=True)),
    train_cfg=dict(),
    test_cfg=dict(
        flip_test=True,
        post_process='unbiased',
        shift_heatmap=True,
        modulate_kernel=11))

data_root = 'data/coco'

data_cfg = dict(
    image_size=[256, 256],
    heatmap_size=[64, 64],
    num_output_channels=channel_cfg['num_output_channels'],
    num_joints=channel_cfg['dataset_joints'],
    dataset_channel=channel_cfg['dataset_channel'],
    inference_channel=channel_cfg['inference_channel'],
)

train_pipeline = [
    # dict(type='TopDownGetBboxCenterScale', padding=1.25),
    dict(type='TopDownRandomFlip', flip_prob=0.5),
    dict(
        type='TopDownGetRandomScaleRotation', rot_factor=30,
        scale_factor=0.25),
    dict(type='TopDownAffine'),
    dict(type='MMToTensor'),
    dict(
        type='NormalizeTensor',
        mean=[0.485, 0.456, 0.406],
        std=[0.229, 0.224, 0.225]),
    dict(type='TopDownGenerateTarget', sigma=3),
    dict(
        type='PoseCollect',
        keys=['img', 'target', 'target_weight'],
        meta_keys=[
            'image_file', 'image_id', 'joints_3d', 'joints_3d_visible',
            'center', 'scale', 'rotation', 'flip_pairs'
        ])
]

val_pipeline = [
    dict(type='TopDownAffine'),
    dict(type='MMToTensor'),
    dict(
        type='NormalizeTensor',
        mean=[0.485, 0.456, 0.406],
        std=[0.229, 0.224, 0.225]),
    dict(
        type='PoseCollect',
        keys=['img'],
        meta_keys=[
            'image_file', 'image_id', 'center', 'scale', 'rotation',
            'flip_pairs'
        ])
]

test_pipeline = val_pipeline
data_source_cfg = dict(type='HandCocoPoseTopDownSource', data_cfg=data_cfg)

data = dict(
    imgs_per_gpu=32,  # for train
    workers_per_gpu=2,  # for train
    # imgs_per_gpu=1,  # for test
    # workers_per_gpu=1,  # for test
    val_dataloader=dict(samples_per_gpu=32),
    test_dataloader=dict(samples_per_gpu=32),
    train=dict(
        type='HandCocoWholeBodyDataset',
        data_source=dict(
            ann_file=f'{data_root}/annotations/coco_wholebody_train_v1.0.json',
            img_prefix=f'{data_root}/train2017/',
            **data_source_cfg),
        pipeline=train_pipeline),
    val=dict(
        type='HandCocoWholeBodyDataset',
        data_source=dict(
            ann_file=f'{data_root}/annotations/coco_wholebody_val_v1.0.json',
            img_prefix=f'{data_root}/val2017/',
            test_mode=True,
            **data_source_cfg),
        pipeline=val_pipeline),
    test=dict(
        type='HandCocoWholeBodyDataset',
        data_source=dict(
            ann_file=f'{data_root}/annotations/coco_wholebody_val_v1.0.json',
            img_prefix=f'{data_root}/val2017/',
            test_mode=True,
            **data_source_cfg),
        pipeline=val_pipeline),
)

eval_config = dict(interval=10, metric='PCK', save_best='PCK')
evaluator_args = dict(
    metric_names=['PCK', 'AUC', 'EPE', 'NME'], pck_thr=0.2, auc_nor=30)
eval_pipelines = [
    dict(
        mode='test',
        data=dict(**data['val'], imgs_per_gpu=1),
        evaluators=[dict(type='KeyPointEvaluator', **evaluator_args)])
]
export = dict(use_jit=False)
checkpoint_sync_export = True
predict = dict(type='HandKeypointsPredictor')



# model settings
model = dict(
    type='Classification',
    train_preprocess=['mixUp'],
    pretrained=False,
    mixup_cfg=dict(
        mixup_alpha=0.8,
        cutmix_alpha=1.0,
        cutmix_minmax=None,
        prob=1.0,
        switch_prob=0.5,
        mode='batch',
        label_smoothing=0.1,
        num_classes=1000),
    backbone=dict(
        type='VisionTransformer',
        img_size=[224],
        num_classes=1000,
        patch_size=16,
        embed_dim=768,
        depth=12,
        num_heads=12,
        mlp_ratio=4,
        qkv_bias=True,
        drop_rate=0.,
        drop_path_rate=0.1),
    head=dict(
        type='ClsHead',
        loss_config={
            'type': 'SoftTargetCrossEntropy',
        },
        with_fc=False))



# Copyright (c) Alibaba, Inc. and its affiliates.
from pathlib import Path

import numpy as np
import torch
from sklearn.decomposition import PCA
from tqdm import tqdm

from easycv.utils.metric_distance import (CosineSimilarity,
                                          DotproductSimilarity, LpDistance)
from .base_evaluator import Evaluator
from .builder import EVALUATORS
from .metric_registry import METRICS


@EVALUATORS.register_module
class RetrivalTopKEvaluator(Evaluator):
    """ RetrivalTopK evaluator,
      Retrival evaluate do the topK retrival, by measuring the distance of every 1 vs other.
      get the topK nearest, and count the match of ID. if Retrival = 1, Miss = 0. Finally average all
      RetrivalRate.
  """

    def __init__(self,
                 topk=(1, 2, 4, 8),
                 norm=0,
                 metric='cos',
                 pca=0,
                 dataset_name=None,
                 metric_names=['R@K=1'],
                 save_results=False,
                 save_results_dir='',
                 feature_keyword=['neck']):
        '''
        Args:
            top_k: tuple of int, evaluate top_k acc
        '''
        self._topk = topk
        self.norm = 0
        self.metric = metric
        self.dataset_name = dataset_name
        self.pca = pca

        self.save_results = save_results
        self.save_results_dir = Path(save_results_dir)
        if self.save_results:
            assert self.save_results_dir != '', 'when save retrival results formatted as .npy, save_results_dir should be set in config'

        self.feature_keyword = feature_keyword
        super(RetrivalTopKEvaluator, self).__init__(dataset_name, metric_names)

    def _evaluate_impl(self, results, gt_label, step=100):
        res = {}
        for key in self.feature_keyword:
            torch.cuda.empty_cache()
            res1 = self._evaluate(
                results, gt_label, step=100, feature_keyword=key)
            res.update(res1)

        return res

    def _evaluate(self, results, gt_label, step=100, feature_keyword='neck'):
        """Retrival evaluate do the topK retrival, by measuring the distance of every 1 vs other.
            get the topK nearest, and count the match of ID. if Retrival = 1, Miss = 0. Finally average all
            RetrivalRate.

        """
        # first print() is to show shape clearly in multi-process situation. don't comment it
        print()

        print('retrieval available keys : ', results.keys())
        results = results[feature_keyword]
        gt_label = gt_label

        if len(results.shape) > 2:
            results = results.view(results.shape[0], -1)

        print("retrieval results' shape:", results.shape)
        print('ground truth label shape:', gt_label.shape)

        if self.pca > 0:
            print(self.pca)
            pca_model = PCA(n_components=self.pca)
            c_results = results.cpu().detach().numpy()
            pca_model.fit(c_results)
            results_numpy = pca_model.transform(c_results)
            results = torch.from_numpy(results_numpy).to(gt_label.device)
            print('After pca')
            print(results.shape)

        assert results.size(0) == gt_label.size(0)

        # add GPU resouce to speed up
        try:
            results = results.cuda()
            gt_label = gt_label.cuda()
        except:
            pass

        if self.norm > 0:
            results = torch.nn.functional.normalize(
                results, p=self.norm, dim=1)
            gt_label = torch.nn.functional.normalize(
                gt_label, p=self.norm, dim=1)

        distance_dict = {
            'cos': CosineSimilarity,
            'dot': DotproductSimilarity,
            'lp': LpDistance,
        }

        distance_mult = {
            'cos': -1,
            'dot': -1,
            'lp': 1,
        }

        distance = distance_dict[self.metric]

        topk_list = self._topk
        topk_res = {}

        if type(topk_list) == int:
            topk_list = [topk_list]

        for k in topk_list:
            topk_res[k] = 0

        if self.save_results:
            retrival_index_results = torch.zeros([1, max(topk_list)],
                                                 dtype=torch.int64).to(
                                                     results.device)
            retrival_distance_results = torch.zeros([1, max(topk_list)],
                                                    dtype=torch.float32).to(
                                                        results.device)
            retrival_topk_results = torch.zeros(
                [results.size(0), len(topk_list)], dtype=torch.int8)

        for start_idx in tqdm(range(0, results.size(0), step)):
            dis_matrix = distance_mult[self.metric] * distance(
                results[start_idx:start_idx + step, ...], results)
            # this diag to eliminate distance with its self, which should be smallest, so we add max to kill it
            diag = torch.ones(dis_matrix.size(0)).to(
                dis_matrix.device) * torch.max(dis_matrix)
            diag = diag.to(dis_matrix.device)
            dis_matrix[range(dis_matrix.size(0)),
                       range(start_idx, start_idx + dis_matrix.size(0))] = diag

            query_distance, query_index = torch.topk(
                dis_matrix, k=max(topk_list), dim=1, largest=False)

            if self.save_results:
                retrival_index_results = torch.cat(
                    (retrival_index_results, query_index), 0)
                retrival_distance_results = torch.cat(
                    (retrival_distance_results, query_distance), 0)

            for idx in range(dis_matrix.size(0)):
                gt_query = torch.index_select(
                    gt_label, dim=0, index=query_index[idx]).cpu().numpy()
                for topk in topk_list:
                    gt_query_k = gt_query[:topk]
                    catch = False
                    for qi in range(topk):
                        if gt_query_k[qi] == gt_label[start_idx + idx]:
                            catch = True
                            if self.save_results:
                                retrival_topk_results[
                                    start_idx + idx,
                                    topk_list.index(topk)] = 1
                            break

                    if catch:
                        topk_res[topk] += 1

        if self.save_results:
            retrival_index_results = retrival_index_results[1:, ].cpu().numpy()
            retrival_distance_results = retrival_distance_results[
                1:, ].cpu().numpy()
            retrival_topk_results = retrival_topk_results.cpu().numpy()
            save_results_dict = {}
            save_results_dict[
                'retrival_index_results'] = retrival_index_results
            save_results_dict[
                'retrival_distance_results'] = retrival_distance_results
            save_results_dict['retrival_topk_results'] = retrival_topk_results
            save_results_dict['gt_label'] = gt_label.cpu().numpy()
            if self.dataset_name is not None:
                file_name = self.save_results_dir / '{}_retrival_results.npy'.format(
                    self.dataset_name)
            else:
                file_name = self.save_results_dir / 'retrival_results.npy'
            np.save(file_name, save_results_dict)

        eval_res = {}

        for k in topk_list:
            key_name = 'R@K={}'.format(k)
            eval_res[key_name] = (float(topk_res[k]) / results.size(0)) * 100

        print("Retrieval Eval of %s 's %s feature Result :" %
              (self.dataset_name, feature_keyword))
        print(eval_res)

        return eval_res


METRICS.register_default_best_metric(RetrivalTopKEvaluator, 'R@K=1', 'max')



# Copyright (c) Alibaba, Inc. and its affiliates.
from .base import BaseGCN
from .skeleton_gcn import SkeletonGCN
from .stgcn_backbone import STGCN
from .stgcn_head import STGCNHead

__all__ = ['BaseGCN', 'SkeletonGCN', 'STGCN', 'STGCNHead']



from easycv.datasets.registry import PIPELINES


@PIPELINES.register_module()
class TextTokenizer:

    def __init__(
        self,
        tokenizer_type='bert-base-chinese',
        max_length=50,
        padding='max_length',
        truncation=True,
    ):
        from transformers import BertTokenizerFast
        self.tokenizer = BertTokenizerFast.from_pretrained(tokenizer_type)
        self.max_length = max_length
        self.padding = padding
        self.truncation = truncation

    def __call__(self, results):
        text = results['text']
        tokens = self.tokenizer(
            text,
            max_length=self.max_length,
            padding=self.padding,
            return_tensors='pt',
            truncation=True)

        results['text_input_ids'] = tokens.input_ids.reshape([-1])
        results['text_input_mask'] = tokens.attention_mask.reshape([-1])
        return results
